#ifndef _TVP_PLUGIN_MT19937AR_H_
#define _TVP_PLUGIN_MT19937AR_H_

#include "TVPPlugIn.h"

/*
   A C-program for MT19937
   C++ wrapped version by W.Dee <dee@kikyou.info>
   TVP Plugin version by mey
*/

#define TP_MT_N 624

struct TPMersenneTwisterData
{
	int left;
	unsigned long *next; /* points a value in 'state' */
	unsigned long state[TP_MT_N]; /* the array for the state vector  */
};

class TPMersenneTwister : protected TPMersenneTwisterData
{

public:
	TPMersenneTwister(unsigned long s = 5489UL);
		/* initializes state[N] with a seed */
	TPMersenneTwister(unsigned long init_key[], unsigned long key_length);
		/* initialize by an array with array-length */
		/* init_key is the array for initializing keys */
		/* key_length is its length */
	TPMersenneTwister(const TPMersenneTwisterData &data);
		/* construct tTJSMersenneTwisterData data */

	virtual ~TPMersenneTwister() {;}

private:
	void init_genrand(unsigned long s);

	void next_state(void);

public:
	unsigned long int32(void); /* generates a random number on [0,0xffffffff]-interval */
	long int31(void); /* generates a random number on [0,0x7fffffff]-interval */
	double real1(void); /* generates a random number on [0,1]-real-interval */
	double real2(void); /* generates a random number on [0,1)-real-interval */
	double real3(void); /* generates a random number on (0,1)-real-interval */
	double res53(void); /* generates a random number on [0,1) with 53-bit resolution*/

	double rand_double(void); /* generates a random number on [0,1) with IEEE 64-bit double precision */

	const TPMersenneTwisterData & GetData() const { return *this; }
		/* retrieve data */
	void SetData(const TPMersenneTwisterData & rhs);
		/* set data */
};


#endif _TVP_PLUGIN_MT19937AR_H_
