#include "TVPPlugIn.h"


// is null?
bool isNullObject(tTJSVariant * obj)
{
	return ((obj->AsObjectClosureNoAddRef()).Object == NULL);
}

bool isNullObject(tTJSVariantClosure * obj)
{
	return (obj->Object == NULL);
}


// instanceof Zq
bool isObjectInstanceOf(iTJSDispatch2 * obj, const tjs_char * name)
{
	static tjs_uint32 hint = 0;
	return (obj->IsInstanceOf(0, NULL, &hint, name, obj) == TJS_S_TRUE);
}

bool isObjectInstanceOf(const tTJSVariantClosure * obj, const tjs_char * name)
{
	static tjs_uint32 hint = 0;
	return (obj->IsInstanceOf(0, NULL, &hint, name, NULL) == TJS_S_TRUE);
}

bool isObjectInstanceOf(const tTJSVariant * var, const tjs_char * name)
{
	static tjs_uint32 hint = 0;
	tTJSVariantClosure * obj = &(var->AsObjectClosureNoAddRef());
	return (obj->IsInstanceOf(0, NULL, &hint, name, NULL) == TJS_S_TRUE);
}



// Array Util.
//------------------------------------------------------------------------------
// create Array
//------------------------------------------------------------------------------
void Array_CreateObject(tTJSVariant * result)
{
	TVPExecuteExpression(TJS_W("new Array();"), result);
}

//------------------------------------------------------------------------------
// property get Array.count
//------------------------------------------------------------------------------
tjs_int Array_PropGet_count(iTJSDispatch2 * obj, const tjs_char * memberName)
{
	tjs_int rv = -1;
	try {
		rv = TJSGetArrayElementCount(obj);
	} catch (...) {
		TP_ThrowMes(TJSEM_E_PROPGETBYNUM_FAIL, memberName, TJS_W("Array.count"));
	}
	return rv;
}

tjs_int Array_PropGet_count(iTJSDispatch2 * obj)
{
	return Array_PropGet_count(obj, TJSEM_ANON_FUNC);
}


// zRs[
void Array_FuncCall_assign(iTJSDispatch2 * src, iTJSDispatch2 * dest)
{
	Array_FuncCall_assign(src, dest, TJSEM_ANON_FUNC);
}

void Array_FuncCall_assign(iTJSDispatch2 * src, iTJSDispatch2 * dest, const tjs_char * memberName)
{
	const num_params = 1;

	static tjs_uint32 hint = 0;
	static tTJSVariant val[num_params] = { NULL };

	val[0] = src;
	tTJSVariant *pval[1] = { val };
	TVP_PLUGIN_TJSDISPATCH2_FUNCCALL_CALLIMPL(dest, Array, assign, &hint, NULL, num_params, pval, memberName);

	return;
}

// Array.erase
void Array_FuncCall_erase(iTJSDispatch2 * src, tjs_int idx)
{
	Array_FuncCall_erase(src, idx, TJSEM_ANON_FUNC);
}

void Array_FuncCall_erase(iTJSDispatch2 * src, tjs_int idx, const tjs_char * memberName)
{
	const num_params = 1;

	static tTJSVariant result;
	static tjs_uint32 hint = 0;
	static tTJSVariant val[num_params] = { (tjs_int64)0 };

	val[0] = (tjs_int64)idx;
	tTJSVariant *pval[1] = {val};
	TVP_PLUGIN_TJSDISPATCH2_FUNCCALL_CALLIMPL(src, Array, erase, &hint, &result, num_params, pval, memberName)
	return;
}


// Array.add
void Array_FuncCall_add(iTJSDispatch2 * src, tTJSVariant * obj)
{
	Array_FuncCall_add(src, obj, TJSEM_ANON_FUNC);
}

void Array_FuncCall_add(iTJSDispatch2 * src, tTJSVariant * obj, const tjs_char * memberName)
{
	const num_params = 1;

	static tTJSVariant result;
	static tjs_uint32 hint = 0;
	static tTJSVariant * val[num_params] = { NULL };

	val[0] = obj;
	TVP_PLUGIN_TJSDISPATCH2_FUNCCALL_CALLIMPL(src, Array, add, &hint, &result, num_params, val, memberName)
	return;
}

// Array.clear
void Array_FuncCall_clear(iTJSDispatch2 * src)
{
	Array_FuncCall_clear(src, TJSEM_ANON_FUNC);
}

void Array_FuncCall_clear(iTJSDispatch2 * src, const tjs_char * memberName)
{
	static tTJSVariant result;
	static tjs_uint32 hint = 0;
	TVP_PLUGIN_TJSDISPATCH2_FUNCCALL_CALLIMPL(src, Array, clear, &hint, &result, 0, NULL, memberName)
	return;
}


//------------------------------------------------------------------------------
// property get (by num) Array[n] as iTJSDispatch2 *
//------------------------------------------------------------------------------
const iTJSDispatch2 * Array_PropGetByNumAsObject(iTJSDispatch2 * obj, tjs_int num, const tjs_char * memberName)
{
	static tTJSVariant workVal;
	if (TJS_FAILED(obj->PropGetByNum(TJS_MEMBERMUSTEXIST, num, &workVal, obj))) {
		TP_ThrowMes(TJSEM_E_PROPGETBYNUM_FAIL, memberName, TJS_W("Array"), (ttstr)num);
	}
	return workVal.AsObjectNoAddRef();
}

//------------------------------------------------------------------------------
// property get (by num) Array[n] as iTJSDispatch2 *
//------------------------------------------------------------------------------
const iTJSDispatch2 * Array_PropGetByNumAsObject(iTJSDispatch2 * obj, tjs_int num)
{
	return Array_PropGetByNumAsObject(obj, num, TJSEM_ANON_FUNC);
}

//------------------------------------------------------------------------------
// property get (by num) Array[n] as tTJSVariant &
//------------------------------------------------------------------------------
const tTJSVariant & Array_PropGetByNumAsVariant(iTJSDispatch2 * obj, tjs_int num, const tjs_char * memberName)
{
	static tTJSVariant workVal;
	if (TJS_FAILED(obj->PropGetByNum(TJS_MEMBERMUSTEXIST, num, &workVal, obj))) {
		TP_ThrowMes(TJSEM_E_PROPGETBYNUM_FAIL, memberName, TJS_W("Array"), (ttstr)num);
	}
	return (tTJSVariant &)workVal;
}

//------------------------------------------------------------------------------
// property get (by num) Array[n] as tTJSVariant &
//------------------------------------------------------------------------------
const tTJSVariant & Array_PropGetByNumAsVariant(iTJSDispatch2 * obj, tjs_int num)
{
	return Array_PropGetByNumAsVariant(obj, num, TJSEM_ANON_FUNC);
}





// X,Y C^[[u
void Array_deinterleaveXY(iTJSDispatch2 * path, tjs_int arraysize, tjs_int * x, tjs_int * y)
{
	tTJSVariant val; // work

	for (tjs_int i = 0; i < arraysize; i++) {
		x[i] = (tTVInteger)Array_PropGetByNumAsVariant(path, 2*i);
		y[i] = (tTVInteger)Array_PropGetByNumAsVariant(path, 2*i+1);
	}
}

void Array_deinterleaveXY(iTJSDispatch2 * path, tjs_int arraysize, tjs_real * x, tjs_real * y)
{
	tTJSVariant val; // work

	for (tjs_int i = 0; i < arraysize; i++) {
		x[i] = (tTVReal)Array_PropGetByNumAsVariant(path, 2*i);
		y[i] = (tTVReal)Array_PropGetByNumAsVariant(path, 2*i+1);
	}
}


// Oo
void TP_AddLog(const tjs_char * msg, const tjs_char * memberName)
{
	ttstr message = TVPFormatMessage(TPM_MESSAGE_HEADER TJS_W("%1: %2"), ttstr(memberName), msg);
	TVPAddLog(message);
}

void TP_AddLog(const tjs_char * msg, const tjs_char * memberName, const ttstr & param1)
{
	TP_AddLog(TVPFormatMessage(msg, param1).c_str(), memberName);
}

void TP_AddLog(const tjs_char * msg, const tjs_char * memberName, const ttstr & param1, const ttstr & param2)
{
	TP_AddLog(TVPFormatMessage(msg, param1, param2).c_str(), memberName);
}


// ObZ[W𓊂
void TP_ThrowMes(const tjs_char * msg, const tjs_char * memberName)
{
	ttstr message = TVPFormatMessage(TPM_MESSAGE_HEADER TJS_W("%1: %2"), ttstr(memberName), msg);
	TVPThrowExceptionMessage(message.c_str());
}

void TP_ThrowMes(const tjs_char * msg, const tjs_char * memberName, const ttstr & param1)
{
	TP_ThrowMes(TVPFormatMessage(msg, param1).c_str(), memberName);
}

void TP_ThrowMes(const tjs_char * msg, const tjs_char * memberName, const ttstr & param1, const ttstr & param2)
{
	TP_ThrowMes(TVPFormatMessage(msg, param1, param2).c_str(), memberName);
}


// ^
const ttstr & TP_GetVariantTypeName(const tTJSVariant * variant)
{
	return TP_GetVariantTypeName(variant->Type());
}

const ttstr & TP_GetVariantTypeName(const tTJSVariant & variant)
{
	return TP_GetVariantTypeName(variant.Type());
}

const ttstr & TP_GetVariantTypeName(tTJSVariantType type)
{
	static const ttstr ST_INT("");
	static const ttstr ST_REAL("");
	static const ttstr ST_STR("");
	static const ttstr ST_OCT("INebg");
	static const ttstr ST_OBJ("IuWFNg");
	static const ttstr ST_VOID("void");
	static const ttstr ST_UNK("s");

	switch (type) {
	case tvtInteger: return ST_INT;
	case tvtReal:    return ST_REAL;
	case tvtString:  return ST_STR;
	case tvtOctet:   return ST_OCT;
	case tvtObject:  return ST_OBJ;
	case tvtVoid:    return ST_VOID;
	default:         return ST_UNK;
	}
}


void TP_AddConstRegisterExpression(ttstr & expression, const ttstr & name, tjs_int value)
{
	expression += "global.";
	expression += name;
	expression += " = ";
	expression += ttstr(value);
	expression += ";\n";
	return;
}

void TP_AddConstDeleteExpression(ttstr & expression, const ttstr & name)
{
	expression += "delete global.";
	expression += name;
	expression += ";\n";
	return;
}



tTJSVariantOctet * AllocVariantOctetFromString(ttstr & workstr, tjs_int mode, const wchar_t * __tp_member_name)
{
	switch (mode) {
	case TP_DUMP_STR_LOCALE      :
	case TP_DUMP_ENDIAN_PLATFORM :
	case TP_DUMP_ENDIAN_LITTLE   :
	case TP_DUMP_ENDIAN_BIG      :
		break;
	default:
		TP_ThrowMes(TPM_E_INVALID_DUMP_MODE, __tp_member_name);
	};


	tTJSVariantOctet * rv = NULL;
	bool swap = false;

	if (mode == TP_DUMP_ENDIAN_LITTLE) {
		swap = (TJS_HOST_IS_LITTLE_ENDIAN) ? false : true;
	}
	if (mode == TP_DUMP_ENDIAN_BIG) {
		swap = (TJS_HOST_IS_LITTLE_ENDIAN) ? true : false;
	}

	// ]ʂAvbgtH[̃[JɈˑϊ
	if (mode == TP_DUMP_STR_LOCALE) {
		// i[
		tjs_int narrow_len = workstr.GetNarrowStrLen();
		if (narrow_len == -1) TP_ThrowMes(TJSEM_E_WIDE_TO_NARROW_FAIL, __tp_member_name);

		char * narrow_str = new char [narrow_len + 1];

		try {
			// ϊ
			workstr.ToNarrowStr(narrow_str, narrow_len);
			rv = TJSAllocVariantOctet((tjs_uint8 *)narrow_str, narrow_len);
		} catch (...) {
			delete [] narrow_str;
			throw;
		}
		delete [] narrow_str;

	} else {
		// TODO?: TQ[gyȀH
		tjs_int len = workstr.GetLen();
		const tjs_int CHARSIZE = sizeof(tjs_char);
		const tjs_int chlen = CHARSIZE * workstr.GetLen();

		if (swap && workstr.GetLen() > 0) {
			const tjs_uint8 * base = (tjs_uint8 *)(workstr.c_str());
			tjs_uint8 * ref = new tjs_uint8[chlen];
			for (tjs_int i = 0; i < len; i++) {
				for (tjs_int j = 0; j < CHARSIZE; j++) {
					ref[i * CHARSIZE + j] = base[i * CHARSIZE + (CHARSIZE - j - 1)];
				}
			}
			rv = TJSAllocVariantOctet(ref, chlen);
			delete[] ref;

		} else {
			const tjs_uint8 * ref = (tjs_uint8 *)(workstr.c_str());
			rv = TJSAllocVariantOctet(ref, chlen);
		}
	}

	return rv;

}

//---------------------------------------------------------------------------

