//////////////////////////////////////////////////////////////////////////
//
//  hl[Op`
//
//  by pastel
//  http://www.roy.hi-ho.ne.jp/pastel/home/VC/delauney.html
//
//  2001/8/10 first version
//  2005/03/07 TVP Plugin version by mey

#include <windows.h>
#include <math.h>
#include "Delaunay.h"


/////////////////////////////////////////////////////////////
//
// TPDelaunay

TPDelaunay::TPDelaunay()
{
	m_px = m_py = NULL;
	m_nPt = m_nPtCapacity = 0;

	m_pTri = NULL;
	m_nTri = m_nTriCapacity = 0;

	Clear();
}

TPDelaunay::~TPDelaunay()
{
	ClearPtArray();
	ClearTriArray();
}

void TPDelaunay::Clear()
{
	InitPtArray();
	InitTriArray();
}

void TPDelaunay::InitPlane(int w, int h)
{
	int ip0, ip1, ip2, ip3;
	TPTriangle *pTri;
	int itri0, itri1;

	Clear();

	ip0 = AddPtToArray(0, 0);
	ip1 = AddPtToArray(w, 0);
	ip2 = AddPtToArray(w, h);
	ip3 = AddPtToArray(0, h);

	pTri = new TPTriangle(ip0, ip1, ip3);
	itri0 = AddTriToArray(pTri);

	pTri = new TPTriangle(ip1, ip2, ip3);
	itri1 = AddTriToArray(pTri);

	SetAdjacent(itri0, itri1);
}

void TPDelaunay::InitQuad(double x0, double y0, double x1, double y1, double x2, double y2, double x3, double y3)
{
	int ip0, ip1, ip2, ip3;
	TPTriangle *pTri;
	int itri0, itri1;

	Clear();

	ip0 = AddPtToArray(x0, y0);
	ip1 = AddPtToArray(x1, y1);
	ip2 = AddPtToArray(x2, y2);
	ip3 = AddPtToArray(x3, y3);

	pTri = new TPTriangle(ip0, ip1, ip3);
	itri0 = AddTriToArray(pTri);

	pTri = new TPTriangle(ip1, ip2, ip3);
	itri1 = AddTriToArray(pTri);

	SetAdjacent(itri0, itri1);
}


TPTriangle * TPDelaunay::GetTriangle(int ip) const
{
	TPTriangle * tri = NULL;
	tri= this->m_pTri[ip];
	return tri;
}

void TPDelaunay::SetTrianglePt()
{
	int n = this->CountTri();
	int ip[3];

	for (int i=0; i<n; i++) {
		this->GetPtOfTri(i, ip[0], ip[1], ip[2]);
		TPTriangle * tri = this->GetTriangle(i);
		for (int j=0; j<3; j++) {
			double x, y;
			this->GetPtPos(ip[j], x, y);
			tri->SetPtPos(j, x, y);
		}
	}
}


bool TPDelaunay::AddPtPos(double x, double y)
{
	int iTri;
	SIntStack stack;
	int iatri0, iatri1, iatri2;
	TPTriangle *pOldTri;
	TPTriangle *tri0, *tri1, *tri2;
	int a, b, c, p;
	int itri0, itri1, itri2;
	int i;
	int tri, adtri;

	// (x, y)Ɋ܂ގOp`𓾂
	iTri = FindTriContains(x, y);
	if (iTri<0) return false;

	// (x, y)V_ƂĒǉ
	p = AddPtToArray(x, y);

	// iTriɗאڂOp`𓾂DiȂ-1ɂȂj
	iatri0 = GetAdjacent(iTri, 0);
	iatri1 = GetAdjacent(iTri, 1);
	iatri2 = GetAdjacent(iTri, 2);

	// ̎Op`3̏Op`ɕ̂ɒu
	pOldTri = m_pTri[iTri];
	pOldTri->GetPt(a, b, c);

	tri0 = new TPTriangle(a, b, p);
	tri1 = new TPTriangle(b, c, p);
	tri2 = new TPTriangle(c, a, p);
	if (!tri0 || !tri1 || !tri2) {
//		ReportToLog("TPDelaunay::AddPtPos(): out of memory\n");
//		Fatal();
		return (-1);
	}
	m_pTri[iTri] = tri0;
	itri0 = iTri;
	itri1 = AddTriToArray(tri1);
	itri2 = AddTriToArray(tri2);
	SetAdjacent(itri0, itri1);
	SetAdjacent(itri1, itri2);
	SetAdjacent(itri2, itri0);
	if (iatri2!=-1) SetAdjacent(itri0, iatri2);
	if (iatri0!=-1) SetAdjacent(itri1, iatri0);
	if (iatri1!=-1) SetAdjacent(itri2, iatri1);
	delete pOldTri;

	// 3̏Op`X^bNɐς
	stack.Push(itri0);
	stack.Push(itri1);
	stack.Push(itri2);

	// X^bNɎOp`ԌJԂ
	while (stack.Pop(tri)) {

		// triɗאڂOp`ɂČJԂ
		for (i=0; i<3; i++) {
			adtri = GetAdjacent(tri, i);
			if (adtri==-1) continue;

			// ̗אڎOp`̂A3̒_ǂpł͂Ȃ̂I
			int ip0, ip1, ip2;
			GetPtOfTri(adtri, ip0, ip1, ip2);
			if ((ip0!=p) && (ip1!=p) && (ip2!=p)) {

				// p3p`̊Oډ~ɂ邩ǂ`FbND
				double x0, y0, x1, y1, x2, y2;
				double cx, cy, d, r2;

				GetPtPos(ip0, x0, y0);
				GetPtPos(ip1, x1, y1);
				GetPtPos(ip2, x2, y2);

				CalcCircumCircle(x0, y0, x1, y1, x2, y2, cx, cy, r2);
				d = (x-cx)*(x-cx)+(y-cy)*(y-cy);

				if (d<r2) {
					// p͊Oډ~̓ɂD
					// triadtrȋΊpւA
					SwapDiagonal(tri, adtri);

					// ɂ2VɃX^bNɐς
					stack.Push(tri);
					stack.Push(adtri);
				}
			}
		}
	}

	return true;
}

int TPDelaunay::CountPt() const
{
	return m_nPt;
}

bool TPDelaunay::GetPtPos(int iPt, double& x, double& y) const
{
	if ((iPt<0) || (iPt>=m_nPt)) {
//		ReportToLog("TPDelaunay::GetPtPos(): index(%d) out of range.", iPt);
		return false;
	}

	x = m_px[iPt];
	y = m_py[iPt];

	return true;
}

int TPDelaunay::CountTri() const
{
	return m_nTri;
}

bool TPDelaunay::GetPtOfTri(int iTri, int& ip0, int& ip1, int& ip2) const
{
	if ((iTri<0) || (iTri>=m_nTri)) {
//		ReportToLog("TPDelaunay::GetPtOfTri(): index(%d) out of range.", iTri);
		return false;
	}

	TPTriangle *pTri = m_pTri[iTri];
	pTri->GetPt(ip0, ip1, ip2);

	return true;
}

bool TPDelaunay::GetPtPosOfTri(int iTri, double& x0, double& y0, double& x1, double& y1, double& x2, double&y2) const
{
	int ip0, ip1, ip2;

	if (!GetPtOfTri(iTri, ip0, ip1, ip2)) return false;

	GetPtPos(ip0, x0, y0);
	GetPtPos(ip1, x1, y1);
	GetPtPos(ip2, x2, y2);

	return true;
}

int TPDelaunay::GetAdjacent(int iTri, int iEdge) const
{
	if ((iTri<0) || (iTri>=m_nTri)) return -1;

	TPTriangle *pTri = m_pTri[iTri];
	return pTri->GetAdjacent(iEdge);
}

bool TPDelaunay::SetAdjacent(int iTri0, int iTri1)
{
	if ((iTri0==iTri1) || 
	    (iTri0<0) || (iTri0>=m_nTri) || 
		(iTri1<0) || (iTri1>=m_nTri)) return false;

	int ip0[3], ip1[3];
	int c, i, j;
	int a[3];
	int d0[2], d1[2];
	int k;
	int ice0, ice1;

	GetPtOfTri(iTri0, ip0[0], ip0[1], ip0[2]);
	GetPtOfTri(iTri1, ip1[0], ip1[1], ip1[2]);

	c = 0;
	for (i=0; i<3; i++) {
		a[i] = -1;
		for (j=0; j<3; j++) {
			if (ip0[i]==ip1[j]) {
				a[i] = j;
				break;
			}
		}
		if (a[i]!=-1) c++;
	}

	// a[i] ... iTri0iԖڂ̓_ƓiTri1̃CfbNXD
	// (ex)
	//  a[0] = 1   iTri00Ԗڂ̓_́AiTri11Ԗڂ̓_Ɠ
	//  a[1] = 2   iTri01Ԗڂ̓_́AiTri12Ԗڂ̓_Ɠ
	//  a[2] = -1  iTri02Ԗڂ̓_ƓiTri1̓_͂Ȃ

	if (c!=2) return false;	// 2̎Op`͗אڕӂĂȂ

	k = 0;
	for (i=0; i<3; i++) {
		if (a[i]!=-1) {
			d0[k] = i;
			d1[k] = a[i];
			k++;
		}
	}

	// d0[0], d0[1]iTri0̋Lӂ̗[_̃CfbNX
	// d1[0], d1[1]iTri1̋Lӂ̗[_̃CfbNX

	switch (d0[0]+d0[1]) {
		case 1: ice0 = 2; break;
		case 3: ice0 = 0; break;
		case 2: ice0 = 1; break;
	}
		
	switch (d1[0]+d1[1]) {
		case 1: ice1 = 2; break;
		case 3: ice1 = 0; break;
		case 2: ice1 = 1; break;
	}
	
	// Lӂ̂ŃZbg
	m_pTri[iTri0]->SetAdjacent(ice0, iTri1);
	m_pTri[iTri1]->SetAdjacent(ice1, iTri0);

	return true;
}

int TPDelaunay::FindTriContains(double x, double y) const
{
	int i;

	for (i=0; i<m_nTri; i++) {
		if (IsInside(i, x, y)) return i;
	}

	return -1;
}

bool TPDelaunay::IsInside(int iTri, double x, double y) const
{
	if ((iTri<0) || (iTri>=m_nTri)) return false;

	int ip0, ip1, ip2;
	double x0, y0, x1, y1, x2, y2;

	GetPtOfTri(iTri, ip0, ip1, ip2);

	GetPtPos(ip0, x0, y0);
	GetPtPos(ip1, x1, y1);
	GetPtPos(ip2, x2, y2);

	if (!IsPlusRotation(x1-x0, y1-y0, x-x0, y-y0)) return false;
	if (!IsPlusRotation(x2-x1, y2-y1, x-x1, y-y1)) return false;
	if (!IsPlusRotation(x0-x2, y0-y2, x-x2, y-y2)) return false;

	return true;
}

bool TPDelaunay::IsAdjacent(int iTri0, int iTri1, int& c0, int& c1) const
{
	int a[3], b[3];
	int i, j;

	for (i=0; i<3; i++) a[i] = GetAdjacent(iTri0, i);
	for (i=0; i<3; i++) b[i] = GetAdjacent(iTri1, i);

	for (i=0; i<3; i++) {
		for (j=0; j<3; j++) {
			if ((a[i]!=-1) && (b[j]!=-1) && (a[i]==iTri1) && (b[j]==iTri0)) {
				c0 = i;
				c1 = j;
				return true;
			}
		}
	}

	return false;
}

bool TPDelaunay::SwapDiagonal(int iTri0, int iTri1)
{
	int ce0, ce1;
	int nce00, nce01, nce10, nce11;
	int ip[3], iq[3];
	int d0, d1;
	int e0, e1;
	int atri00, atri01, atri10, atri11;

	if (!IsAdjacent(iTri0, iTri1, ce0, ce1)) return false;
	// ce0 ... tri0̋Lӂ̔ԍ
	// ce1 ... tri1̋Lӂ̔ԍ
	// Lӂ̔ԍ́A̕ӂ̌ɂ_̔ԍƓ̂
	// Op`3̓_̂ALӂ̔ԍƓȂ2̓_
	// Lӂ̗[̓_ɂȂD
	
	GetPtOfTri(iTri0, ip[0], ip[1], ip[2]);
	GetPtOfTri(iTri1, iq[0], iq[1], iq[2]);

	e0 = ip[ce0];
	e1 = iq[ce1];
	// (e0, e1) ... ʕӂłȂ̑Ίp̗[_̔ԍ

	nce00 = nce01 = -1;
	switch (ce0) {
		case 0: nce00 = 1; nce01 = 2; break;
		case 1: nce00 = 2; nce01 = 0; break;
		case 2: nce00 = 0; nce01 = 1; break;
	}

	nce10 = nce11 = -1;
	switch (ce1) {
		case 0: nce10 = 1; nce11 = 2; break;
		case 1: nce10 = 2; nce11 = 0; break;
		case 2: nce10 = 0; nce11 = 1; break;
	}
	// nce00, nce01
	//   tri0̋LӂłȂ2̕ӂ̔ԍ
	//   i͋Lӂ̗[_̔ԍƓj
	//   ȂAKnce00, nce01̏ŔvɂȂ
	// nce10, nce11
	//   tri1ɑ΂LƓӖ̒lD

	d0 = ip[nce00];
	d1 = ip[nce01];
	// (d0, d1) ... ʕӂ̗[_̔ԍ

	// Ƃ̎Op`̗אڎOp`𓾂
	atri00 = GetAdjacent(iTri0, nce00);
	atri01 = GetAdjacent(iTri0, nce01);
	atri10 = GetAdjacent(iTri1, nce10);
	atri11 = GetAdjacent(iTri1, nce11);

	// VOp`ɒu
	if (m_pTri[iTri0]) delete m_pTri[iTri0];
	m_pTri[iTri0] = new TPTriangle(e1, e0, d0);
	if (!m_pTri[iTri0]) {
//		ReportToLog("TPDelaunay::SwapDiagonal(): out of memory");
		return false;
	}

	if (m_pTri[iTri1]) delete m_pTri[iTri1];
	m_pTri[iTri1] = new TPTriangle(e0, e1, d1);
	if (!m_pTri[iTri1]) {
//		ReportToLog("TPDelaunay::SwapDiagonal(): out of memory");
		return false;
	}

	// אڏXV
	SetAdjacent(iTri0, iTri1);

	if (atri01!=-1) SetAdjacent(iTri0, atri01);
	if (atri10!=-1) SetAdjacent(iTri0, atri10);

	if (atri00!=-1) SetAdjacent(iTri1, atri00);
	if (atri11!=-1) SetAdjacent(iTri1, atri11);

	return true;
}

int TPDelaunay::CalcCircumCircle(
	double x0, double y0, double x1, double y1, double x2, double y2,
	double& cx, double& cy, double& r2)
{
	double hx0 = (x0+x1)/2;
	double hy0 = (y0+y1)/2;
	double hx1 = (x1+x2)/2;
	double hy1 = (y1+y2)/2;
	double dx0 = -(y1-y0);
	double dy0 = x1-x0;
	double dx1 = -(y2-y1);
	double dy1 = x2-x1;
	double t, det;
	
	det = dx0*(-dy1)-(-dx1)*dy0;
	if (fabs(det)<0.000001) {
//		ReportToLog("Error: deteminant is nearly zero\n");
//		Fatal();
		return (-6);
	}

	t = ((-dy1)*(hx1-hx0)+dx1*(hy1-hy0))/det;

	// SW	
	cx = hx0+t*dx0;
	cy = hy0+t*dy0;

	// a̓
	r2 = (x0-cx)*(x0-cx)+(y0-cy)*(y0-cy);

	return 1;
}

bool TPDelaunay::IsPlusRotation(double x0, double y0, double x1, double y1) const
{
	double z2 = x0*y1 - x1*y0;
	
	return z2 > 0 ? true : false;
}

int TPDelaunay::AddPtToArray(double x, double y)
{
	if (m_nPt>=m_nPtCapacity) {
		if (ExpandPtArray() == -1) return -1;
	}
	
	m_px[m_nPt] = x;
	m_py[m_nPt] = y;

	return m_nPt++;
}

int TPDelaunay::AddTriToArray(TPTriangle *pTri)
{
	if (m_nTri>=m_nTriCapacity) {
		if (ExpandTriArray() == -1) return -1;
	}

	m_pTri[m_nTri] = pTri;

	return m_nTri++;
}

void TPDelaunay::ClearPtArray()
{
	if (m_px) {
		free(m_px);
		m_px = NULL;
	}

	if (m_py) {
		free(m_py);
		m_py = NULL;
	}

	m_nPt = 0;
	m_nPtCapacity = 0;
}

int TPDelaunay::InitPtArray()
{
	ClearPtArray();

	m_px = (double *)calloc(DELAUNAY_INIT_PTSIZE, sizeof(double));
	if (!m_px) {
//		ReportToLog("[%s, %d] malloc failed", __FILE__, __LINE__);
		return -1;
	}

	m_py = (double *)calloc(DELAUNAY_INIT_PTSIZE, sizeof(double));
	if (!m_py) {
//		ReportToLog("[%s, %d] malloc failed", __FILE__, __LINE__);
		return -1;
	}

	m_nPt = 0;
	m_nPtCapacity = DELAUNAY_INIT_PTSIZE;

	return m_nPtCapacity;
}

int TPDelaunay::ExpandPtArray()
{
	int newSize = m_nPtCapacity * 2;
	int i;

	m_px = (double *)realloc(m_px, sizeof(double)*newSize);
	if (!m_px) {
//		ReportToLog("[%s, %d] realloc failed", __FILE__, __LINE__);
		return -1;
	}

	m_py = (double *)realloc(m_py, sizeof(double)*newSize);
	if (!m_py) {
//		ReportToLog("[%s, %d] realloc failed", __FILE__, __LINE__);
		return -1;
	}

	for (i=m_nPt; i<newSize; i++) {
		m_px[i] = 0.0;
		m_py[i] = 0.0;
	}

	m_nPtCapacity = newSize;

	return m_nPtCapacity;
}

void TPDelaunay::ClearTriArray()
{
	int i;
	TPTriangle *pTri;

	if (m_pTri) {
		for (i=0; i<m_nTriCapacity; i++) {
			pTri = m_pTri[i];
			if (pTri) delete pTri;
		}

		free(m_pTri);
		m_pTri = NULL;
	}

	m_nTri = 0;
	m_nTriCapacity = 0;
}

int TPDelaunay::InitTriArray()
{
	ClearTriArray();

	m_pTri = (TPTriangle **)calloc(DELAUNAY_INIT_TRISIZE, sizeof(TPTriangle *));
	if (!m_pTri) {
//		ReportToLog("[%s, %d] malloc failed", __FILE__, __LINE__);
		return -1;
	}
	
	m_nTri = 0;
	m_nTriCapacity = DELAUNAY_INIT_TRISIZE;

	return m_nTriCapacity;
}

int TPDelaunay::ExpandTriArray()
{
	int newSize = m_nTriCapacity * 2;
	int i;

	m_pTri = (TPTriangle **)realloc(m_pTri, sizeof(TPTriangle *)*newSize);
	if (!m_pTri) {
//		ReportToLog("[%s, %d] realloc failed", __FILE__, __LINE__);
		return -1;
	}

	for (i=m_nTri; i<newSize; i++) m_pTri[i] = NULL;

	m_nTriCapacity = newSize;
	return m_nTriCapacity;
}


/////////////////////////////////////////////////////////////
//
//  TPTriangle

TPTriangle::TPTriangle()
{
	m_iPt[0] = m_iPt[1] = m_iPt[2] = -1;
	m_iAdjacent[0] = m_iAdjacent[1] = m_iAdjacent[2] = -1;
}

TPTriangle::TPTriangle(int ip0, int ip1, int ip2)
{
	m_iPt[0] = ip0;
	m_iPt[1] = ip1;
	m_iPt[2] = ip2;

	m_iAdjacent[0] = m_iAdjacent[1] = m_iAdjacent[2] = -1;
}

int TPTriangle::SetAdjacent(int iEdge, int iAdjacent)
{
	if ((iEdge<0) || (iEdge>2)) {
//		ReportToLog("TPDelaunay::TPTriangle::SetAdjacent(): edge index(%d) is out of range.", iEdge);
//		Fatal();
		return (-3);
	}

	m_iAdjacent[iEdge] = iAdjacent;
	return 1;
}

int TPTriangle::GetAdjacent(int iEdge) const
{
	if ((iEdge<0) || (iEdge>2)) {
//		ReportToLog("TPDelaunay::TPTriangle::GetAdjacent(): edge index(%d) is out of range.", iEdge);
//		Fatal();
		return -2;
	}

	return m_iAdjacent[iEdge];
}

void TPTriangle::GetPt(int& ip0, int& ip1, int& ip2) const
{
	ip0 = m_iPt[0];
	ip1 = m_iPt[1];
	ip2 = m_iPt[2];
}

void TPTriangle::GetPtPos(int ip, double& px, double& py) const
{
	px = m_X[ip];
	py = m_Y[ip];
}

void TPTriangle::SetPtPos(int ip, double px, double py)
{
	m_X[ip] = px;
	m_Y[ip] = py;
}


/////////////////////////////////////////////////////////////
//
// SIntStack

TPDelaunay::SIntStack::SIntStack()
{
	m_pValue = NULL;
	m_Count = m_Capacity = 0;

	InitStack();
}

TPDelaunay::SIntStack::~SIntStack()
{
	ClearStack();
}

int TPDelaunay::SIntStack::Push(int x)
{
	if (m_Count>=m_Capacity) {
		if (!ExpandStack()) {
//			ReportToLog("SIntStack::Push(): expand stack failed\n");
//			Fatal();
			return -7;
		}
	}

	m_pValue[m_Count++] = x;
	return x;
}

bool TPDelaunay::SIntStack::Pop(int &x)
{
	if (m_Count>0) {
		x = m_pValue[m_Count-1];
		m_Count--;
		return true;
	}

	return false;
}

int TPDelaunay::SIntStack::Count() const
{
	return m_Count;
}

void TPDelaunay::SIntStack::ClearStack()
{
	if (m_pValue) {
		free(m_pValue);
		m_pValue = NULL;
	}

	m_Count = m_Capacity = 0;
}

void TPDelaunay::SIntStack::InitStack()
{
	ClearStack();

	m_pValue = (int *)calloc(DELAUNAY_INIT_STACKSIZE, sizeof(int));
	if (!m_pValue) {
//		ReportToLog("SIntStack::InitStack(): failed to alloc memory\n");
//		Fatal();
		return ;
	}

	m_Count = 0;
	m_Capacity = DELAUNAY_INIT_STACKSIZE;
}

bool TPDelaunay::SIntStack::ExpandStack()
{
	int newSize = m_Capacity * 2;
	int i;

	m_pValue = (int *)realloc(m_pValue, newSize*sizeof(int));
	if (!m_pValue) {
//		ReportToLog("SIntStack::ExpandStack(): failed to realloc memory\n");
		return false;
	}
	
	for (i=m_Count; i<newSize; i++) {
		m_pValue[i] = 0;
	}

	m_Capacity = newSize;
	return true;
}


