#include "Morph.h"
#include "TPGraphic.h"

void TrianglePos::getPosOnEdge(unsigned int sp, unsigned int ep, int ratio, int& x, int& y) const
{
	if (sp <= 2 && ep <= 2) {
		x = ux[sp] + ((ux[ep] - ux[sp]) >> 16) * ratio;
		y = uy[sp] + ((uy[ep] - uy[sp]) >> 16) * ratio;
	}
}


TrianglePos::TrianglePos()
{
	set(0, 0, 0, 0, 0, 0);
}

TrianglePos::TrianglePos(TPTriangle * tri)
{
	double x0, y0, x1, y1, x2, y2;
	tri->GetPtPos(0, x0, y0);
	tri->GetPtPos(1, x1, y1);
	tri->GetPtPos(2, x2, y2);
	set(x0, y0, x1, y1, x2, y2);
}

TrianglePos::TrianglePos(double x0, double y0, double x1, double y1, double x2, double y2)
{
	set(x0, y0, x1, y1, x2, y2);
}

void TrianglePos::set(double x0, double y0, double x1, double y1, double x2, double y2)
{
#define DBL_TO_FPINT16(a) (((int)(a)) << 16)
	sx[0] = DBL_TO_FPINT16(x0);
	sy[0] = DBL_TO_FPINT16(y0);
	sx[1] = DBL_TO_FPINT16(x1);
	sy[1] = DBL_TO_FPINT16(y1);
	sx[2] = DBL_TO_FPINT16(x2);
	sy[2] = DBL_TO_FPINT16(y2);
	copy();
#undef DBL_TO_FPINT16
}

void TrianglePos::copy()
{
	for (int i=0;i<3;i++) {
		ux[i] = sx[i];
		uy[i] = sy[i];
	}
	order[0] = 0;
	order[1] = 1;
	order[2] = 2;
}

bool TrianglePos::orderBy(const TrianglePos * ref)
{
	return orderBy(ref->order[0], ref->order[1], ref->order[2]);
}

bool TrianglePos::orderBy(unsigned int a0, unsigned int a1, unsigned int a2)
{
	bool rv = false;
	if ((a0 <= 2) && (a1 <= 2) && (a2 <= 2) && (a0 != a1) && (a0 != a2) && (a1 != a2)) {
		ux[0] = sx[a0];
		uy[0] = sy[a0];
		ux[1] = sx[a1];
		uy[1] = sy[a1];
		ux[2] = sx[a2];
		uy[2] = sy[a2];
		order[0] = a0;
		order[1] = a1;
		order[2] = a2;
		rv = true;
	}
	return rv;
}

void TrianglePos::orderByY()
{
	copy();
	int tmp;
#define SWAP_INT(a,b,t) t=a;a=b;b=t;
	if (uy[0] > uy[1]) {
		SWAP_INT(uy[0], uy[1], tmp);
		SWAP_INT(ux[0], ux[1], tmp);
		SWAP_INT(order[0], order[1], tmp);
	}
	if (uy[0] > uy[2]) {
		SWAP_INT(uy[0], uy[2], tmp);
		SWAP_INT(ux[0], ux[2], tmp);
		SWAP_INT(order[0], order[2], tmp);
	}
	if (uy[1] > uy[2]) {
		SWAP_INT(uy[1], uy[2], tmp);
		SWAP_INT(ux[1], ux[2], tmp);
		SWAP_INT(order[1], order[2], tmp);
	}
#undef SWAP_INT
}

// SɓĂtrue
bool TrianglePos::isContentOf(const tTVPRect * rect) const
{
	bool rv = true;
	for (int i = 0; i < 3; i++) {
		int x = ux[i] >> 16;
		int y = uy[i] >> 16;
		if (x < rect->left || x >= rect->right || y < rect->top || y >= rect->bottom) {
			rv = false;
			break;
		}
	}
	return rv;
}

// SɓĂtrue
bool TrianglePos::isContentOf(tjs_int width, tjs_int height) const
{
	bool rv = true;
	for (int i = 0; i < 3; i++) {
		int x = ux[i] >> 16;
		int y = uy[i] >> 16;
		if (x < 0 || x >= width || y < 0 || y >= height) {
			rv = false;
			break;
		}
	}
	return rv;
}




//---------------------------------------------------------------------------
// Morph
//---------------------------------------------------------------------------
tjs_int TPFUNCCALL morphTriangleImpl(
	const ImageBufferInfo * destInfo,
	const ImageBufferInfo * src1Info,
	const ImageBufferInfo * src2Info,
	TrianglePos * destTri,
	TrianglePos * src1Tri,
	TrianglePos * src2Tri,
	tjs_real morphRatio)
{
	// Y\[g
	destTri->orderByY();
	// dest\[g
	src1Tri->orderBy(destTri);
	src2Tri->orderBy(destTri);

	tjs_int opa = morphRatio * 255;
//	tjs_int opa = 0;

	tjs_uint32 * src1Bufferptr = (tjs_uint32 * )src1Info->bufferptr;
	tjs_uint32 * src2Bufferptr = (tjs_uint32 * )src2Info->bufferptr;
	tjs_int src1Bufferpitch = src1Info->bufferpitch;
	tjs_int src2Bufferpitch = src2Info->bufferpitch;

	// raster scan

	// X^XLKvissNZiYj
	int dyPixel01 = (destTri->uy[1] - destTri->uy[0]) >> 16;
	int dyPixel02 = (destTri->uy[2] - destTri->uy[0]) >> 16;
	int dyPixel12 = (destTri->uy[2] - destTri->uy[1]) >> 16;

	// X^XLissNZiYj
	int progress = 0;

	// destOp`㕔X^XL
	while (progress <= dyPixel01) {
		// X^XLis
		int ratio01 = dyPixel01 ? ((progress << 16) / dyPixel01) : 0x8000;
		int ratio02 = dyPixel02 ? ((progress << 16) / dyPixel02) : 0x8000;

		// O̎Op`̕ӏ̓_
		// signed 1.15.16bit Œ菬_
		int s1xs, s1ys, s1xe, s1ye;
		int s2xs, s2ys, s2xe, s2ye;
		int dxs, dys, dxe, dye;

		destTri->getPosOnEdge(0, 1, ratio01, dxs, dys);
		destTri->getPosOnEdge(0, 2, ratio02, dxe, dye);
		src1Tri->getPosOnEdge(0, 1, ratio01, s1xs, s1ys);
		src1Tri->getPosOnEdge(0, 2, ratio02, s1xe, s1ye);
		src2Tri->getPosOnEdge(0, 1, ratio01, s2xs, s2ys);
		src2Tri->getPosOnEdge(0, 2, ratio02, s2xe, s2ye);

		// Rs[FX^[ (>0)
		int destDX = (dxe - dxs) >> 16;
		int destWidthAbs = (destDX < 0) ? -destDX : destDX;
		destWidthAbs++;

		// Rs[FXgXLXebv
		// signed 1.15.16bit Œ菬_
		int s1stepx = (s1xe - s1xs) / destWidthAbs;
		int s1stepy = (s1ye - s1ys) / destWidthAbs;
		int s2stepx = (s2xe - s2xs) / destWidthAbs;
		int s2stepy = (s2ye - s2ys) / destWidthAbs;

		tjs_uint32 * destBufferptr;
		if (destDX >= 0) {
			destBufferptr = destInfo->getBufferPtr(dxs >> 16, (destTri->uy[0] >> 16) + progress);
			TVPPlugInLinTransConstAlphaBlend_SD(
				destBufferptr, destWidthAbs, 
				src1Bufferptr, s1xs, s1ys, s1stepx, s1stepy, src1Bufferpitch,
				src2Bufferptr, s2xs, s2ys, s2stepx, s2stepy, src2Bufferpitch,
				opa);
		} else {
			destBufferptr = destInfo->getBufferPtr(dxe >> 16, (destTri->uy[0] >> 16) + progress);
			TVPPlugInLinTransConstAlphaBlend_SD(
				destBufferptr, destWidthAbs, 
				src1Bufferptr, s1xe, s1ye, -s1stepx, -s1stepy, src1Bufferpitch,
				src2Bufferptr, s2xe, s2ye, -s2stepx, -s2stepy, src2Bufferpitch,
				opa);
		}

		progress++;
	}


	progress = 1;
	// destOp`X^XL
	while (progress <= dyPixel12) {
		// is
		int ratio12 = dyPixel12 ? ((progress << 16) / dyPixel12) : 0x8000;
		int ratio02 = dyPixel02 ? (((progress + dyPixel01) << 16) / dyPixel02) : 0x8000;

		// O̎Op`̕ӏ̓_
		int s1xs, s1ys, s1xe, s1ye;
		int s2xs, s2ys, s2xe, s2ye;
		int dxs, dys, dxe, dye;
		destTri->getPosOnEdge(1, 2, ratio12, dxs, dys);
		destTri->getPosOnEdge(0, 2, ratio02, dxe, dye);
		src1Tri->getPosOnEdge(1, 2, ratio12, s1xs, s1ys);
		src1Tri->getPosOnEdge(0, 2, ratio02, s1xe, s1ye);
		src2Tri->getPosOnEdge(1, 2, ratio12, s2xs, s2ys);
		src2Tri->getPosOnEdge(0, 2, ratio02, s2xe, s2ye);

		// Rs[FX^[ (>0)
		int destDX = (dxe - dxs) >> 16;
		int destWidthAbs = (destDX < 0) ? -destDX : destDX;
		destWidthAbs++;

		// Rs[FXgXLXebv
		int s1stepx = (s1xe - s1xs) / destWidthAbs;
		int s1stepy = (s1ye - s1ys) / destWidthAbs;
		int s2stepx = (s2xe - s2xs) / destWidthAbs;
		int s2stepy = (s2ye - s2ys) / destWidthAbs;

		tjs_uint32 * destBufferptr;
		if (destDX >= 0) {
			destBufferptr = destInfo->getBufferPtr(dxs >> 16, (destTri->uy[1] >> 16) + progress);
			TVPPlugInLinTransConstAlphaBlend_SD(
				destBufferptr, destWidthAbs, 
				src1Bufferptr, s1xs, s1ys, s1stepx, s1stepy, src1Bufferpitch,
				src2Bufferptr, s2xs, s2ys, s2stepx, s2stepy, src2Bufferpitch,
				opa);
		} else {
			destBufferptr = destInfo->getBufferPtr(dxe >> 16, (destTri->uy[1] >> 16) + progress);
			TVPPlugInLinTransConstAlphaBlend_SD(
				destBufferptr, destWidthAbs, 
				src1Bufferptr, s1xe, s1ye, -s1stepx, -s1stepy, src1Bufferpitch,
				src2Bufferptr, s2xe, s2ye, -s2stepx, -s2stepy, src2Bufferpitch,
				opa);
		}

		progress++;
	}

	return opa;
}


tjs_int TPFUNCCALL morphImpl(
	const ImageBufferInfo * destInfo,
	const ImageBufferInfo * src1Info,
	const ImageBufferInfo * src2Info,
	const tjs_real * points1x,
	const tjs_real * points1y,
	const tjs_real * points2x,
	const tjs_real * points2y,
	tjs_int nPoints,
	tjs_real morphRatio)
{
	TPDelaunay dt;
	dt.InitQuad(
		points1x[0], points1y[0], 
		points1x[1], points1y[1], 
		points1x[2], points1y[2], 
		points1x[3], points1y[3]);

	for (tjs_int i = 4; i < nPoints; i++) {
		dt.AddPtPos(points1x[i], points1y[i]);
	}

	return morphImpl(&dt, destInfo, src1Info, src2Info, 
		points1x, points1y, points2x, points2y, nPoints, morphRatio);
}



tjs_int TPFUNCCALL morphImpl(
	const TPDelaunay * triangulator,
	const ImageBufferInfo * destInfo,
	const ImageBufferInfo * src1Info,
	const ImageBufferInfo * src2Info,
	const tjs_real * points1x,
	const tjs_real * points1y,
	const tjs_real * points2x,
	const tjs_real * points2y,
	tjs_int nPoints,
	tjs_real morphRatio)
{
	tjs_int nTriangle = triangulator->CountTri();
	for (tjs_int i = 0; i < nTriangle; i++) {

		// src1
		double x1[3], y1[3];
		triangulator->GetPtPosOfTri(i, x1[0], y1[0], x1[1], y1[1], x1[2], y1[2]);

		// src2
		int ip[3];
		double x2[3], y2[3];
		triangulator->GetPtOfTri(i, ip[0], ip[1], ip[2]);
		for (tjs_int j = 0; j < 3; j++) {
			x2[j] = points2x[ip[j]];
			y2[j] = points2y[ip[j]];
		}

		// Op`_
		TrianglePos src1Tri(x1[0], y1[0], x1[1], y1[1], x1[2], y1[2]);
		TrianglePos src2Tri(x2[0], y2[0], x2[1], y2[1], x2[2], y2[2]);
		TrianglePos destTri(
			x1[0] + (x2[0] - x1[0]) * morphRatio,
			y1[0] + (y2[0] - y1[0]) * morphRatio,
			x1[1] + (x2[1] - x1[1]) * morphRatio,
			y1[1] + (y2[1] - y1[1]) * morphRatio,
			x1[2] + (x2[2] - x1[2]) * morphRatio,
			y1[2] + (y2[2] - y1[2]) * morphRatio);
/*
		// ͂ݏo`FbN
		if (!src1Tri.isContentOf(src1Info->width, src1Info->height)) {
			TP_ThrowMes(TPM_E_OUT_OF_RECT, TPM_ANON_FUNC);
		}
		if (!src2Tri.isContentOf(src2Info->width, src2Info->height)) {
			TP_ThrowMes(TPM_E_OUT_OF_RECT, TPM_ANON_FUNC);
		}
*/
		morphTriangleImpl(
			destInfo, src1Info, src2Info,
			&destTri, &src1Tri, &src2Tri,
			morphRatio);
	}

	return 1;

}

