/*
 * PunycodeGR[_/fR[_
 *
 * - RFC3492 QlɁAPunycodeGR[_/fR[_̂łB
 * - NAMEPREP(RFC3491)̓e͊܂܂܂BȂ킿ÃNX͓o͕
 *   IDNAxƂĂ̐؂܂B
 *
 * ̃\[XR[hɊւSĂ̒쌠܂B
 * oO̕񍐂́A k-tak at letter.or.jp ܂ŁB
 */

/*
 * ǋLF̃R[h
 *   C : Punycode
 *   http://www.void.in/wiki/Punycode
 * ŌJĂ̂łB
 * 
 * 2005-11-21 mey
 */


#ifndef ___PUNYCODE_H___
#define ___PUNYCODE_H___

#include <string>
#include <vector>
#include <algorithm>
#include <functional>
#include <climits>



class Punycode
{
   public:
      static const unsigned int PUNYCODE_BASE		= 36;
      static const unsigned int PUNYCODE_TMIN		= 1;
      static const unsigned int PUNYCODE_TMAX		= 26;
      static const unsigned int PUNYCODE_SKEW		= 38;
      static const unsigned int PUNYCODE_DAMP		= 700;
      static const unsigned int PUNYCODE_INITIAL_BIAS	= 72;
      static const unsigned int PUNYCODE_INITIAL_N	= 128;
      
   private:
      Punycode()
      {}
      
      Punycode(const Punycode&)
      {}
      
   public:
      /**
       * PunycodeUTF8ɃfR[hB
       * @param punycode [in] Punycode
       * @return UTF8Bs΋󕶎
       * @note punycodexn--prefixĂȂꍇAASCIIƔfĂ̂܂ܕԂB
       */
      static std::string decode(const char* punycode)
      {
	 if (!punycode || !punycode[0])	    return std::string();
	 
	 if (punycode[0] && (punycode[0]=='x' || punycode[0]=='X') &&
	     punycode[1] && (punycode[1]=='n' || punycode[1]=='N') &&
	     punycode[2] && punycode[2] == '-' &&
	     punycode[3] && punycode[3] == '-')
	 {
	    // ACE prefixXLbv
	    punycode += 4;
	 } else {
	    // ACE prefixȂBASCIIƔfB
	    return std::string(punycode);
	 }
	 
	 std::string ret;
	 std::vector<unsigned int> codepoints;
	 std::vector<unsigned int>::iterator cpinsit;
	 
	 const char* it = 0;
	 const char* itbegin = 0;
	 const char* itend = 0;
	 
	 const char* lastsep = strrchr(punycode, '-');
	 itend = punycode + strlen(punycode);
	 
	 if (lastsep) {
	    for (const char* p=punycode; p!=lastsep; ++p) {
	       if (*(unsigned char*)p >= 0x80) {
		  // non-basic chars before sep.
		  return std::string();
	       }
	       codepoints.push_back((unsigned int)*(unsigned char*)p);
	    }
	    itbegin = lastsep + 1;
	 } else {
	    // lastsepNULL̏ꍇASĂ̕{R[h|Cgō\Ă邩A
	    // 邢͑S{R[h|Cg܂܂ĂȂ̂ꂩłB
	    itbegin = punycode;
	 }
	 
	 if (itbegin == itend) {
	    if (utf32ToUtf8(ret, codepoints)) {
	       return ret;
	    } else {
	       return std::string();
	    }
	 }
	 
	 unsigned int n, i, oldi, bias, w, k, digit, t, outlen, ndiff;
	 
	 n = PUNYCODE_INITIAL_N;
	 i = 0;
	 bias = PUNYCODE_INITIAL_BIAS;
	 it = itbegin;
	 
	 while (it!=itend) {
	    oldi = i;
	    w = 1;
	    for (k=PUNYCODE_BASE; ; k+=PUNYCODE_BASE) {
	       
	       if (it==itend)		return std::string();
	       digit = decodeDigit((unsigned int)(unsigned char)*it);
	       ++it;
	       
	       if (!calcMulAdd(i, i, digit, w)) {
		  // overflow
		  return std::string();
	       }
	       
	       if (k <= bias) {
		  t = PUNYCODE_TMIN;
	       } else if (k >= bias + PUNYCODE_TMAX) {
		  t = PUNYCODE_TMAX;
	       } else {
		  t = k - bias;
	       }
	       
	       if (digit < t)		break;
	       
	       if (!calcMulAdd(w, 0, w, (PUNYCODE_BASE-t))) {
		  // overflow
		  return std::string();
	       }
	    }
	    
	    outlen = (unsigned int)codepoints.size()+1;
	    bias = adapt(i-oldi, outlen, oldi==0);
	    ndiff = i / outlen;
	    if (UINT_MAX - ndiff < n) {
	       // overflow
	       return std::string();
	    }
	    n += ndiff;
	    i = i % outlen;
	    if (n < 0x80) {
	       return std::string();
	    }
	    cpinsit = codepoints.begin();
	    cpinsit += i;
	    codepoints.insert(cpinsit, n);
	    ++i;
	 }
	 
	 if (utf32ToUtf8(ret, codepoints)) {
	    return ret;
	 } else {
	    return std::string();
	 }
      }
      
      /**
       * UTF8PunycodeɃGR[hB
       * @param utf8 [in] GR[hΏۂUTF-8B
       * @return PunycodeBsƋ󕶎B
       * @note utf8ASCIÎ݂ō\ꍇÂ܂ܕԂB
       */
      static std::string encode(const char* utf8)
      {
	 std::string ret;
	 
	 std::vector<unsigned int> codepoints;
	 std::vector<unsigned int>::const_iterator it, itend;
	 unsigned int numpoints;
	 
	 if (!utf8ToUtf32(codepoints, utf8))	return std::string();
	 numpoints = (unsigned int)codepoints.size();
	 itend = codepoints.end();
	 
	 /////////////////////////////////////////
	 // encoding start
	 unsigned int m, n, delta, bias, h, b, q, k, t; // same meaning in RFC3492
	 
	 b = 0;
	 for (it=codepoints.begin(); it!=itend; ++it) {
	    // {R[h|CgȂ炻̂܂܃Rs[
	    if (*it < 0x80) {
	       ret.push_back((char)(unsigned char)(*it));
	       b++;
	    }
	 }
	 if (b > 0 && b < numpoints) {
	    ret.push_back('-');
	 }
	 
	 if (b == numpoints) {
	    // ͊̕{R[h|Cĝ݂ō\ĂB
	    // GR[h̕KvȂB
	    return ret;
	 }
	 
	 n = PUNYCODE_INITIAL_N;
	 delta = 0;
	 bias = PUNYCODE_INITIAL_BIAS;
	 h = b;
	 
	 while (h < numpoints) {
	    // nȏ̒lR[h|CĝAŏ̂̂TB
	    m = UINT_MAX;
	    for (it=codepoints.begin(); it!=itend; ++it) {
	       if (m > *it && *it >= n) {
		  m = *it;
	       }
	    }
	    
	    // deltaZo
	    if (!calcMulAdd(delta, delta, (m-n), (h+1))) {
	       // overflow
	       return std::string();
	    }
	    n = m;
	    
	    for (it=codepoints.begin(); it!=itend; ++it) {
	       if (*it < n) {
		  if (delta >= UINT_MAX) {
		     return std::string(); //overflow
		  }
		  delta++;
	       } else if (*it == n) {
		  q = delta;
		  for (k=PUNYCODE_BASE; ; k+=PUNYCODE_BASE) {
		     if (k <= bias) {
			t = PUNYCODE_TMIN;
		     } else if (k >= (bias + PUNYCODE_TMAX)) {
			t = PUNYCODE_TMAX;
		     } else {
			t = (k - bias);
		     }
		     
		     if (q < t)	    break;
		     ret.push_back(encodeDigit(t + ((q-t) % (PUNYCODE_BASE-t))));
		     q = (q-t) / (PUNYCODE_BASE-t);
		  }
		  ret.push_back(encodeDigit(q));
		  
		  bias = adapt(delta, h+1, h==b);
		  delta = 0;
		  h++;
	       }
	    }
	    delta++;
	    n++;
	 }
	 
	 return "xn--" + ret;
      }
      
   protected:
      /**
       * BOMȂUTF8(UTF8-N)AUTF32R[h|CgɕϊB
       *
       * @param codepoints [out] ϊꂽR[h|Cgi[B
       * @param utf8 [in] ϊΏە
       * @return trueBsȃV[PX݂ĊSɕϊłȂꍇfalseB
       */
      static bool utf8ToUtf32(std::vector<unsigned int>& codepoints, const char* utf8)
      {
	 if (!utf8 || !utf8[0])		return false;
	 
	 unsigned char* p;
	 unsigned int cp;
	 unsigned int byteval;
	 int cprestbytes;

	 cprestbytes = 0;
	 for (p=(unsigned char*)utf8; *p; p++) {
	    byteval = *p;
	    if (byteval & 0x80) {
	       // 1xxxxxxx
	       if (cprestbytes > 0) {
		  // trailing bytes
		  if ((byteval & 0xC0) == 0x80) {
		     // 10xxxxxx ... trailing bytes
		     cp = (cp << 6) | (byteval & 0x3F);
		     cprestbytes--;
		     if (cprestbytes <= 0) {
			codepoints.push_back(cp);
		     }
		  } else {
		     return false;
		  }
	       } else {
		  // leading byte
		  if ((byteval & 0xE0) == 0xC0) {
		     // 110xxxxx
		     cp = (byteval & 0x1F);
		     cprestbytes = 1;
		  } else if ((byteval & 0xF0) == 0xE0) {
		     // 1110xxxx
		     cp = (byteval & 0x0F);
		     cprestbytes = 2;
		  } else if ((byteval & 0xF8) == 0xF0) {
		     // 11110xxx
		     cp = (byteval & 0x07);
		     cprestbytes = 3;
		  } else if ((byteval & 0xFC) == 0xF8) {
		     // 111110xx
		     cp = (byteval & 0x03);
		     cprestbytes = 4;
		  } else if ((byteval & 0xFE) == 0xFC) {
		     // 1111110x
		     cp = (byteval & 0x01);
		     cprestbytes = 5;
		  } else {
		     // error.
		     return false;
		  }
	       }
	    } else {
	       if (cprestbytes <= 0) {
		  // codepoint 0x00-0x7f
		  codepoints.push_back((unsigned int)*p);
	       } else {
		  // error
		  return false;
	       }
	    }
	 }

	 return (cprestbytes == 0);
      }

      /**
       * UnicodeR[h|CgAUTF-8ɕϊB
       * @param utf8 [out] ϊꂽBOMȂUTF-8񂪊i[
       * @param codepoints [in] UnicodeR[h|Cg
       * @return trueAsfalse
       */
      static bool utf32ToUtf8(std::string& utf8, const std::vector<unsigned int>& codepoints)
      {
	 std::vector<unsigned int>::const_iterator it;
	 std::vector<unsigned int>::const_iterator itend = codepoints.end();

	 unsigned int cp;

	 for (it=codepoints.begin(); it!=itend; ++it) {
	    cp = *it;

	    if (cp < 0x80) {
	       // ASCII codepoint. pass it through.
	       utf8.push_back((char)(unsigned char)cp);

	    } else if (cp < 0x800) {
	       // 2byte sequence
	       // 110xxxxx 10xxxxxx
	       utf8.push_back((char)(unsigned char)(((cp & 0x07C0) >> 6) | 0xC0));
	       utf8.push_back((char)(unsigned char)((cp & 0x3F) | 0x80));

	    } else if (cp < 0xD800) {
	       // first half of 3byte sequence
	       // 1110xxxx 10xxxxxx 10xxxxxx
	       utf8.push_back((char)(unsigned char)(((cp & 0xF000) >> 12) | 0xE0));
	       utf8.push_back((char)(unsigned char)(((cp & 0xFC0) >> 6) | 0x80));
	       utf8.push_back((char)(unsigned char)((cp & 0x3F) | 0x80));

	    } else if (cp < 0xE000) {
	       // surrogate code point. error.
	       return false;
	    } else if (cp < 0x10000) {
	       // latter half of 3byte sequence
	       utf8.push_back((char)(unsigned char)(((cp & 0xF000) >> 12) | 0xE0));
	       utf8.push_back((char)(unsigned char)(((cp & 0xFC0) >> 6) | 0x80));
	       utf8.push_back((char)(unsigned char)((cp & 0x3F) | 0x80));

	    } else if (cp < 0x100000) {
	       // 4byte sequence
	       // 11110xxx 10xxxxxx 10xxxxxx 10xxxxxx
	       utf8.push_back((char)(unsigned char)(((cp & 0x1C0000) >> 18) | 0xF0));
	       utf8.push_back((char)(unsigned char)(((cp & 0x3F000) >> 12) | 0x80));
	       utf8.push_back((char)(unsigned char)(((cp & 0xFC0) >> 6) | 0x80));
	       utf8.push_back((char)(unsigned char)((cp & 0x3F) | 0x80));

	    } else if (cp < 0x200000) {
	       // 5byte sequence
	       // 111110xx 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx
	       utf8.push_back((char)(unsigned char)(((cp & 0x3000000) >> 24) | 0xF8));
	       utf8.push_back((char)(unsigned char)(((cp & 0xFC0000) >> 18) | 0x80));
	       utf8.push_back((char)(unsigned char)(((cp & 0x3F000) >> 12) | 0x80));
	       utf8.push_back((char)(unsigned char)(((cp & 0xFC0) >> 6) | 0x80));
	       utf8.push_back((char)(unsigned char)((cp & 0x3F) | 0x80));

	    } else if (cp < 0x80000000) {
	       // 6byte sequence
	       // 1111110x 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx
	       utf8.push_back((char)(unsigned char)(((cp & 0x40000000) >> 30) | 0xFC));
	       utf8.push_back((char)(unsigned char)(((cp & 0x3F000000) >> 24) | 0x80));
	       utf8.push_back((char)(unsigned char)(((cp & 0xFC0000) >> 18) | 0x80));
	       utf8.push_back((char)(unsigned char)(((cp & 0x3F000) >> 12) | 0x80));
	       utf8.push_back((char)(unsigned char)(((cp & 0xFC0) >> 6) | 0x80));
	       utf8.push_back((char)(unsigned char)((cp & 0x3F) | 0x80));

	    } else {
	       // ȏUTF8ŕ\ȂB
	       return false;
	    }
	 }

	 return true;
      }

      static unsigned int adapt(unsigned int delta, unsigned int numpoints, bool firsttime)
      {
	 unsigned int k, deltamin;

	 if (firsttime) {
	    delta /= PUNYCODE_DAMP;
	 } else {
	    delta /= 2;
	 }
	 delta += (delta / numpoints);

	 k = 0;
	 deltamin = ((PUNYCODE_BASE - PUNYCODE_TMIN) * PUNYCODE_TMAX) / 2;
	 while (delta > deltamin) {
	    delta /= (PUNYCODE_BASE - PUNYCODE_TMIN);
	    k += PUNYCODE_BASE;
	 }

	 return (k + (((PUNYCODE_BASE-PUNYCODE_TMIN+1)*delta) / (delta+PUNYCODE_SKEW)));
      }

      // calc (a + b * c).
      // return true if no overflow occured. else false.
      static bool calcMulAdd(
	 unsigned int& res,
	 unsigned int a, unsigned int b, unsigned int c)
      {
	 if (b > ((UINT_MAX - a) / c)) {
	    return false;
	 }

	 res = a + b * c;
	 return true;
      }

      static char encodeDigit(unsigned int val)
      {
	 return (char)(unsigned char)((val + 22 + 75*(val<26)) & 0xff);
      }

      static unsigned int decodeDigit(unsigned int cp)
      {
	 return (cp-48 < 10 ? cp-22 : cp-65 < 26 ? cp-65 : cp-97<26 ? cp-97 : PUNYCODE_BASE);
      }
};


#endif
