// isqlite.cpp : R\[ AvP[ṼGg |Cg`܂B
//

#include "isqlite.h"


// Chϊ̒擾B
static int GetWideLength(const char * s)
{
	wchar_t * dummy = NULL;
	int buflen = MultiByteToWideChar(932, 0, s, -1, dummy, 0);
	return buflen;
}

// ChɕϊB
static void SetWideChar(const char * s, wchar_t * p, int len)
{
	MultiByteToWideChar(932, 0, s, -1, p, len);
}

// i[ϊ̒擾B
static int GetNarrowLength(const wchar_t * s)
{
	char * dummy = NULL;
	int buflen = WideCharToMultiByte(CP_ACP, 0, s, -1, dummy, 0, NULL, NULL);
	return buflen;
}

// i[ɕϊB
static void SetNarrowChar(const wchar_t * s, char * p, int len)
{
	WideCharToMultiByte(CP_ACP, 0, s, -1, p, len, NULL, NULL);
}




// Xg[ɗ
// binary
static void Put(FILE *stream, const char * str, size_t len)
{
	if (len > 0) {
		fwrite(str, sizeof(char), len, stream);
	}
	return;
}

// Xg[ɗ
// narrow
static void Put(FILE *stream, const char * str)
{
	int len = ::strlen(str);
	if (len > 0) {
		fwrite(str, sizeof(char), len, stream);
	}
	return;
}

// Xg[ɗ
// wide
static void Put(FILE *stream, const wchar_t * str)
{
	int len = GetNarrowLength(str);
	if (len > 0) {
		char *nc = new char[len];
		SetNarrowChar(str, nc, len);
		if (nc[len-1] == '\0') len--;	//kI[H
		fwrite(nc, sizeof(char), len, stream);
		if (nc) delete[] nc;
	}
	return;
}

// Wo̓Xg[ɗ
// narrow
static void Put(const char * str)
{
	Put(DEFAULT_OUT_STREAM, str);
	return;
}

// Wo̓Xg[ɗ
// wide
static void Put(const wchar_t * str)
{
	Put(DEFAULT_OUT_STREAM, str);
	return;
}



// printfbp[
// narrow
static void Print(FILE *stream, const char * format, ...)
{
	int len = ::strlen(format);
	if (len > 0) {
		va_list ap;
		va_start(ap, format);
		vfprintf(stream, format, ap);
		va_end(ap);
	}
	return;
}

// printfbp[
// wide
static void Print(FILE *stream, const wchar_t * format, ...)
{
	int len = GetNarrowLength(format);
	if (len > 0) {
		va_list ap;
		va_start(ap, format);
		char *nc = new char[len+1];
		SetNarrowChar(format, nc, len);
		nc[len] = '\0';
		vfprintf(stream, nc, ap);
		if (nc) delete[] nc;
		va_end(ap);
	}
	return;
}


// printfbp[
// narrow
static void Print(const char * format, ...)
{
	int len = ::strlen(format);
	if (len > 0) {
		va_list ap;
		va_start(ap, format);
		vfprintf(DEFAULT_OUT_STREAM, format, ap);
		va_end(ap);
	}
	return;
}

// printfbp[
// wide
static void Print(const wchar_t * format, ...)
{
	int len = GetNarrowLength(format);
	if (len > 0) {
		va_list ap;
		va_start(ap, format);
		char *nc = new char[len+1];
		SetNarrowChar(format, nc, len);
		nc[len] = '\0';
		vfprintf(DEFAULT_OUT_STREAM, nc, ap);
		if (nc) delete[] nc;
		va_end(ap);
	}
	return;
}






using namespace sqlite3plus;

class Args
{
public:
	SqliteString db;
	SqliteString sql;

	bool batch;

public:
	Args() : batch(false)
	{
	}

	~Args()
	{
	}

	void usage()
	{
		Put(SQ3M_I_USAGE);
	}

	bool parse(int argc, _TCHAR* argv[])
	{
		db.clear();
		sql.clear();

		for(int i=1; i<argc; i++)
		{
			if(_tcscmp(_T("-b"), argv[i])==0)
			{
				batch = true;
			}
			else
			{
				if(db.empty())
				{
					db = argv[i];
				}
				else
				{
					if(!sql.empty()) sql.append(_T(" "));
					sql.append(argv[i]);
				}
			}
		}

		return !db.empty();
	}
};


SqliteString trimLine(SqliteString& str)
{
	size_t posStart = str.find_first_not_of(_T(" \t\r\n"));
	if (posStart == str.npos) return SqliteString();

	size_t posEnd = str.find_last_not_of(_T(" \t\r\n"));
	return str.substr(posStart, posEnd-posStart+1);
}


SqliteString firstToken(SqliteString& str)
{
	size_t posStart = str.find_first_of(_T(" \t\r\n"));
	if(posStart != str.npos)
	{
		return str.substr(0, posStart);
	}
	else
	{
		return str;
	}
}


SqliteString readLine(bool contd, bool quiet)
{
	SqliteString ret;
	CHAR buf[1024];

	if(!quiet)
	{
		(contd) ? Put(SQ3M_I_PROMPT2) : Put(SQ3M_I_PROMPT1);
		fflush(DEFAULT_OUT_STREAM);
	}

	while(fgets(buf, sizeof(buf), stdin))
	{
//		printf("read str : %s\n", buf);
		SqliteString str = SqliteString::fromANSI(buf);
//		Print(_T("converted str : %s\n"), str.c_str());
		ret.append(str.c_str());

		size_t len = strlen(buf);
		if(len > 0)
		{
			if(buf[len-1] == '\n')
			{
				break;
			}
		}
	}

	if(ret.empty()) throw SqliteException("End of input.");

	return ret;
}

SqliteString checkSpecials(SqliteString& sql)
{
	SqliteString cmd = firstToken(trimLine(sql));

	if( cmd.compare(_T(".help"))    == 0 ||
		cmd.compare(_T(".tables"))  == 0 ||
		cmd.compare(_T(".views"))   == 0 ||
		cmd.compare(_T(".indices")) == 0 ||
		cmd.compare(_T(".indexes")) == 0 ||
		cmd.compare(_T(".ddl"))     == 0 ||
		cmd.compare(_T(".columns")) == 0 )
	{
		return cmd;
	}

	return SqliteString();
}


void executeSql(SqliteConnection& con,
				SqliteString& sql,
				bool quiet)
{
	// Translate special commands
	SqliteString cmd = trimLine(sql);

	if(cmd.compare(_T(".help"))==0)
	{
		Put(SQ3M_I_HELP);
		return;
	}
	else if(cmd.compare(_T(".tables"))==0)
	{
		sql = _T("SELECT name FROM sqlite_master WHERE type='table';");
	}
	else if(cmd.compare(_T(".views"))==0)
	{
		sql = _T("SELECT name FROM sqlite_master WHERE type='view';");
	}
	else if(cmd.compare(_T(".indices"))==0 || cmd.compare(_T(".indexes"))==0)
	{
		sql = _T("SELECT name FROM sqlite_master WHERE type='index';");
	}
	else
	{
		SqliteString token = firstToken(cmd);

		if(token.compare(_T(".ddl"))==0)
		{
			sql = _T("SELECT sql FROM sqlite_master WHERE name='");
			sql.append(trimLine(SqliteString(cmd.substr(token.length()))));
			sql.append(_T("';"));
		}
		else if(token.compare(_T(".columns"))==0)
		{
			if(!quiet)
			{
				try
				{
					SqliteString tableName
						= trimLine(SqliteString(cmd.substr(token.length())));
					std::vector<SqliteString> colNames = con.columnNames(tableName.c_str());
					for(size_t i=0; i<colNames.size(); i++)
					{
						if(i != 0) Put(_T(", "));
						Put(_T("'"));
						Put(colNames.at(i).c_str());
						Put(_T("'"));
					}
					Put(_T("\n"));
				}
				catch(SqliteException& e)
				{
					PRINT_EXCEPTION(e);
				}
			}
			return;
		}
	}

	try
	{
		SqliteStatementPtr st = con.createStatement(sql.c_str());
		if(quiet)
		{
			st->execute();
		}
		else
		{
			//NG̊{
			// 1. SqliteStatement::executeQuery()Ăяo
			// 2. SqliteCursorPtrnext()AfalseԂ܂ŌĂяo
			// 3. SqliteCursorPtŕAXR[vAEgƎIɃ\[X
			SqliteCursorPtr result = st->executeQuery();
			while(result->next())
			{
				// 擾o
				for (int i = 0; i < result->colNum(); i++)
				{
					SqliteString val;
					if (result->isNull(i))
					{
						val = _T("NULL");
					}
					else
					{
						val = _T("'");
						val += result->valueAsStr(i) + _T("'");
					}

					if (i != 0) Put(_T(", "));
					Put(val.c_str());
				}

				Put(_T("\n"));
			}

			SqliteString verb = firstToken(cmd);
			if(0 == _tcsicmp(verb.c_str(), _T("select")))
			{
				int selected = result->rowNum();
				Print(SQ3M_I_SELECT, selected);
			}
			else if(
				(0 == _tcsicmp(verb.c_str(), _T("delete"))) ||
				(0 == _tcsicmp(verb.c_str(), _T("insert"))) ||
				(0 == _tcsicmp(verb.c_str(), _T("update")))
				)
			{
				int changed = con.lastChangedRowNum();
				Print(SQ3M_I_UPDATE, changed);
			}
		}
	}
	catch(SqliteException& e)
	{
		PRINT_EXCEPTION(e);
	}
}

void interactiveSQL(SqliteConnection& con, Args& args)
{
	Put(SQ3M_I_HELP_SHORT);
	Put(_T("\n"));

	SqliteString sql;
	while(true)
	{
		try
		{
			SqliteString tmp = trimLine(readLine(!sql.empty(), args.batch));
			if(!sql.empty())
				sql.append(_T(" "));
			sql.append(tmp);
		}
		catch(SqliteException&)
		{
			// thrown by readLine()

			if(!sql.empty())
			{
				Put(_T("Multi-line SQL input aborted.\n"));
				sql.clear();
				continue;
			}
			else
			{
				break;
			}
		}

		// check that sql is complete
		if( !sql.empty() &&
			(!checkSpecials(sql).empty() || con.isCompleteSql(sql.c_str())))
		{
			executeSql(con, sql, args.batch);
			sql.clear();
		}
	}
}



int _tmain(int argc, _TCHAR* argv[])
{
	setlocale(LC_ALL, "");

	Args args;
	if(args.parse(argc, argv))
	{
		// create connection
		SqliteConnection con(args.db.c_str());

		if(args.sql.compare(_T("-"))==0 || args.sql.empty())
		{
			// read sql from stdin
			interactiveSQL(con, args);
		}
		else
		{
			// one liner SQL
			executeSql(con, trimLine(args.sql), args.batch);
		}
	}
	else
	{
		args.usage();
		return 1;
	}

	return 0;
}
