/**************************************************************
*
*  Jul. 8, 2004
*
*  The author disclaims copyright to this source code.
*  In place of a legal notice, here is a blessing:
*
*    May you do good and not evil.
*    May you find forgiveness for yourself and forgive others.
*    May you share freely, never taking more than you give.
*
**************************************************************/

#include "SqliteFunction.h"

namespace sqlite3plus {

/// [U`֐/AOQ[g̃x[XƂȂNX

/// ی삳ꂽRXgN^B̃NX𒼐ڃCX^X邱Ƃ͖B
SqliteFunctionBase::SqliteFunctionBase(sqlite3* db)
	: m_db(db)
	, m_ctx(NULL)
	, m_userData(NULL)
	, m_nArgs(0)
{}

SqliteFunctionBase::~SqliteFunctionBase()
{}

/// ֐ɓnꂽ̃^CvANULLł邩`FbN
bool SqliteFunctionBase::isNull(sqlite3_value* param)
{
	return (sqlite3_value_type(param) == SQLITE_NULL);
}

/// ֐ɓnꂽ̃^CvAł邩`FbN
bool SqliteFunctionBase::isStr(sqlite3_value* param)
{
	return (sqlite3_value_type(param) == SQLITE_TEXT);
}

/// ֐ɓnꂽ̃^CvAł邩`FbN
bool SqliteFunctionBase::isInt(sqlite3_value* param)
{
	return (sqlite3_value_type(param) == SQLITE_INTEGER);
}

/// ֐ɓnꂽ̃^CvAł邩`FbN
bool SqliteFunctionBase::isReal(sqlite3_value* param)
{
	return (sqlite3_value_type(param) == SQLITE_FLOAT);
}

/// ֐ɓnꂽ̃^CvABLOBł邩`FbN
bool SqliteFunctionBase::isBlob(sqlite3_value* param)
{
	return (sqlite3_value_type(param) == SQLITE_BLOB);
}

/// ֐ɓnꂽ̒lAƂĎ擾
SqliteString SqliteFunctionBase::valueAsStr(sqlite3_value* param)
{
	return SqliteString::fromUTF16(
		(const wchar_t*)sqlite3_value_text16(param));
}

/// ֐ɓnꂽ̒lAƂĎ擾
int SqliteFunctionBase::valueAsInt(sqlite3_value* param)
{
	return sqlite3_value_int(param);
}

/// ֐ɓnꂽ̒lA64bitƂĎ擾
__int64 SqliteFunctionBase::valueAsInt64(sqlite3_value* param)
{
	return sqlite3_value_int64(param);
}

/// ֐ɓnꂽ̒lAƂĎ擾
double SqliteFunctionBase::valueAsReal(sqlite3_value* param)
{
	return sqlite3_value_double(param);
}

/// ֐ɓnꂽ̒lABLOBƂĎ擾
const void* SqliteFunctionBase::valueAsBlob(sqlite3_value* param)
{
	return sqlite3_value_blob(param);
}

/// ֐ɓnꂽ̃oCg擾
int SqliteFunctionBase::valueBytes(sqlite3_value* param)
{
	return sqlite3_value_bytes(param);
}

/// ֐ɓnꂽ̃oCg擾
int SqliteFunctionBase::valueBytes16(sqlite3_value* param)
{
	return sqlite3_value_bytes16(param);
}

/// ֐̖߂liljZbg
void SqliteFunctionBase::setReturn(int val)
{
	sqlite3_result_int(m_ctx, val);
}

/// ֐̖߂li64bitljZbg
void SqliteFunctionBase::setReturn(__int64 val)
{
	sqlite3_result_int64(m_ctx, val);
}

/// ֐̖߂liljZbg
void SqliteFunctionBase::setReturn(double val)
{
	sqlite3_result_double(m_ctx, val);
}

/// ֐̖߂lijZbg
void SqliteFunctionBase::setReturn(const _TCHAR* val)
{
	SqliteStringW str = SqliteString(val).toUTF16();
	sqlite3_result_text16(
		m_ctx, str.c_str(),
		(int)str.length()*sizeof(wchar_t),
		SQLITE_TRANSIENT);
}

/// ֐̖߂liBLOBjZbg
void SqliteFunctionBase::setReturn(const void* val, int nBytes, bool makeCopy)
{
	sqlite3_result_blob(m_ctx, val, nBytes,
		makeCopy ? SQLITE_TRANSIENT : SQLITE_STATIC);
}


/// [U`֐쐬邽߂̃x[XNX

/// [U`֐SQLiteɓo^
void SqliteFunction::create(const _TCHAR* funcName, int nArgs, void* userData)
{
	m_name = funcName;
	m_nArgs = nArgs;
	SqliteStringW name = m_name.toUTF16();
	m_userData = userData;

	int res = sqlite3_create_function16(
		m_db, name.c_str(), nArgs, SQLITE_UTF16, this,
		delegateCall, NULL, NULL);

	if(SQLITE_OK != res)
	{
		std::string msg = "Can't create function: ";
		msg += SqliteString(funcName).toANSI().c_str();
		throw ConnectionException(msg.c_str(), res);
	}
}

/// [U`֐̓o^
void SqliteFunction::destroy()
{
	if(!m_name.empty())
	{
		SqliteStringW name = m_name.toUTF16();
		sqlite3_create_function16(
			m_db, name.c_str(), m_nArgs, SQLITE_UTF16, this,
			NULL, NULL, NULL);
	}
}

/// gp
/*static*/ void SqliteFunction::delegateCall(
	sqlite3_context* ctx,
	int nArgs,
	sqlite3_value** params)
{
	SqliteFunction* pThis = reinterpret_cast<SqliteFunction*>(sqlite3_user_data(ctx));
	pThis->m_ctx = ctx;
	pThis->onCall(nArgs, params, pThis->m_userData);
}


/// [U`AOQ[g쐬邽߂̃x[XNX

/// AOQ[gSQLiteɓo^
void SqliteAggregate::create(const _TCHAR* funcName, int nArgs, void* userData)
{
	m_name = funcName;
	m_nArgs = nArgs;
	SqliteStringW name = m_name.toUTF16();
	m_userData = userData;

	int res = sqlite3_create_function16(
		m_db, name.c_str(), nArgs, SQLITE_UTF16, this,
		NULL, delegateStep, delegateFinal);

	if(SQLITE_OK != res)
	{
		std::string msg = "Can't create aggregate: ";
		msg += SqliteString(funcName).toANSI().c_str();
		throw ConnectionException(msg.c_str(), res);
	}
}

/// AOQ[g̓o^
void SqliteAggregate::destroy()
{
	if(!m_name.empty())
	{
		SqliteStringW name = m_name.toUTF16();
		sqlite3_create_function16(
			m_db, name.c_str(), m_nArgs, SQLITE_UTF16, this,
			NULL, NULL, NULL);
	}
}


/// gp
/*static*/ void SqliteAggregate::delegateStep(
	sqlite3_context* ctx,
	int nArgs,
	sqlite3_value** params)
{
	SqliteAggregate* pThis = reinterpret_cast<SqliteAggregate*>(sqlite3_user_data(ctx));
	pThis->onStep(nArgs, params, pThis->m_userData);
}

/// gp
/*static*/ void SqliteAggregate::delegateFinal(
	sqlite3_context* ctx)
{
	SqliteAggregate* pThis = reinterpret_cast<SqliteAggregate*>(sqlite3_user_data(ctx));
	pThis->onFinal(pThis->m_userData);
}


/// [U`r֐쐬邽߂̃x[XNX

/// r֐SQLiteɓo^
void SqliteCollation::create(const _TCHAR* funcName, void* userData)
{
	m_name = funcName;
	m_userData = userData;

	SqliteStringW name = m_name.toUTF16();
	sqlite3_create_collation16(
		m_db, (const char*)name.c_str(), SQLITE_UTF16, this,
		delegateCompare);
}

/// r֐̓o^
void SqliteCollation::destroy()
{
	if(!m_name.empty())
	{
		SqliteStringW name = m_name.toUTF16();
		sqlite3_create_collation16(
			m_db, (const char*)name.c_str(), SQLITE_UTF16, this,
			NULL);
	}
}

/// gp
/*static*/ int SqliteCollation::delegateCompare(
	void* userData,
	int nLen1, const void* str1,
	int nLen2, const void* str2)
{
	SqliteCollation* pThis = reinterpret_cast<SqliteCollation*>(userData);
	SqliteStringW var1, var2;
	var1.assign((const wchar_t*)str1, (size_t)nLen1/sizeof(wchar_t));
	var2.assign((const wchar_t*)str2, (size_t)nLen2/sizeof(wchar_t));

	return pThis->onCompare(
		SqliteString::fromUTF16(var1.c_str()),
		SqliteString::fromUTF16(var2.c_str()),
		pThis->m_userData);
}


}; // end of namespace sqlite3plus

