/**************************************************************
*
*  Jul. 8, 2004
*
*  The author disclaims copyright to this source code.
*  In place of a legal notice, here is a blessing:
*
*    May you do good and not evil.
*    May you find forgiveness for yourself and forgive others.
*    May you share freely, never taking more than you give.
*
**************************************************************/

#include "SqliteStatement.h"

namespace sqlite3plus {

/// J[\

/// f[^x[Xnh擾
sqlite3* SqliteCursor::db()
{
	return m_db;
}

/// Xe[ggnh擾
sqlite3_stmt* SqliteCursor::stmt()
{
	return m_stmt;
}

/// RXgN^B[U͒ڃJ[\NX\zĂ͂ȂȂB
SqliteCursor::SqliteCursor(sqlite3_stmt* stmt, sqlite3* db)
	: m_db(db)
	, m_stmt(stmt)
	, m_rowNum(0)
	, m_eor(false)
{
	m_colNum = sqlite3_column_count(m_stmt);
}

///fXgN^
SqliteCursor::~SqliteCursor()
{}


/// UgZbg̑SĂ̍s擾IĂ邩ǂ𔻒
bool SqliteCursor::eor()
{
	return m_eor;
}

/// ݂܂łɎ擾sԂ
int SqliteCursor::rowNum()
{
	return m_rowNum;
}

/// UgZbg̗񐔂Ԃ
int SqliteCursor::colNum()
{
	return m_colNum;
}

/// ̍s擾
bool SqliteCursor::next()
{
	int res = sqlite3_step(m_stmt);

	if(SQLITE_ROW == res)
	{
		m_rowNum++;
		return true;
	}
	else if(SQLITE_DONE == res)
	{
		m_eor = true;
		return false;
	}
	else
	{
		checkResult(res);
		return false;
	}
}

/// J[\ʒuďAUgZbg̐擪ɖ߂
void SqliteCursor::reset()
{
	int res = sqlite3_reset(m_stmt);
	checkResult(res);
}

/// ݃|CgĂs̎w̒lNULLǂ肷
bool SqliteCursor::isNull(int index)
{
	checkColIndex(index, "isNull");
	return (sqlite3_column_type(m_stmt, index) == SQLITE_NULL);
}

/// ݃|CgĂs̎w̃^Cv񂩂ǂ肷
bool SqliteCursor::isStr(int index)
{
	checkColIndex(index, "isStr");
	return (sqlite3_column_type(m_stmt, index) == SQLITE_TEXT);
}

/// ݃|CgĂs̎w̃^Cvǂ肷
bool SqliteCursor::isInt(int index)
{
	checkColIndex(index, "isInt");
	return (sqlite3_column_type(m_stmt, index) == SQLITE_INTEGER);
}

/// ݃|CgĂs̎w̃^Cvǂ肷
bool SqliteCursor::isReal(int index)
{
	checkColIndex(index, "isReal");
	return (sqlite3_column_type(m_stmt, index) == SQLITE_FLOAT);
}

/// ݃|CgĂs̎w̃^CvBLOBǂ肷
bool SqliteCursor::isBlob(int index)
{
	checkColIndex(index, "isBlob");
	return (sqlite3_column_type(m_stmt, index) == SQLITE_BLOB);
}

/// ݃|CgĂs̎w̒l𕶎ƂĎ擾
SqliteString SqliteCursor::valueAsStr(int index)
{
	checkColIndex(index, "valueAsStr");
	return SqliteString::fromUTF16(
		(const wchar_t*)sqlite3_column_text16(m_stmt, index));
}

/// ݃|CgĂs̎w̒l𐮐lƂĎ擾
int SqliteCursor::valueAsInt(int index)
{
	checkColIndex(index, "valueAsInt");
	return sqlite3_column_int(m_stmt, index);
}

/// ݃|CgĂs̎w̒l64bitlƂĎ擾
__int64 SqliteCursor::valueAsInt64(int index)
{
	checkColIndex(index, "valueAsInt64");
	return sqlite3_column_int64(m_stmt, index);
}

/// ݃|CgĂs̎w̒llƂĎ擾
double SqliteCursor::valueAsReal(int index)
{
	checkColIndex(index, "valueAsReal");
	return sqlite3_column_double(m_stmt, index);
}

/// BLOBl̐擪AhXԂ
const void* SqliteCursor::valueAsBlob(int index)
{
	checkColIndex(index, "valueAsBlob");
	return sqlite3_column_blob(m_stmt, index);
}

/// l߂oCgTCYԂB
int SqliteCursor::valueBytes(int index)
{
	checkColIndex(index, "valueBytes");
	return sqlite3_column_bytes(m_stmt, index);
}

/// l߂oCgTCYԂB
int SqliteCursor::valueBytes16(int index)
{
	checkColIndex(index, "valueBytes16");
	return sqlite3_column_bytes16(m_stmt, index);
}

/// w̖O擾
SqliteString SqliteCursor::colName(int index)
{
	checkColIndex(index, "colName");
	return SqliteString::fromUTF16(
		(const wchar_t*)sqlite3_column_name16(m_stmt, index));
}

void SqliteCursor::checkResult(int rescode)
{
	if(SQLITE_OK != rescode)
	{
		SqliteStringW errMsg = (const wchar_t*)sqlite3_errmsg16(m_db);
		throw QueryException(errMsg.toANSI().c_str(), rescode);
	}
}

void SqliteCursor::checkColIndex(int index, const char* reason)
{
	if(index < 0 || index >= m_colNum)
	{
		throw OutOfBoundsException(reason);
	}
}


/// Xe[gg

/// ێSQL擾
SqliteString SqliteStatement::sql()
{
	return m_sql;
}

/// Xe[ggnh擾
sqlite3_stmt* SqliteStatement::stmt()
{
	return m_stmt;
}

/// f[^x[Xnh擾
sqlite3* SqliteStatement::db()
{
	return m_db;
}

/// RXgN^BSqliteConnection݂̂p
SqliteStatement::SqliteStatement(const _TCHAR* sql, sqlite3_stmt* stmt, sqlite3* db)
	: m_sql(sql)
	, m_stmt(stmt)
	, m_db(db)
{}

SqliteStatement::~SqliteStatement()
{
	if(m_stmt)
		sqlite3_finalize(m_stmt);
}

/// xexecute, executeQueryXe[ggAĂexecuteł悤ɏԂZbg
void SqliteStatement::reset()
{
	int res = sqlite3_reset(m_stmt);
	checkResult(res);
}

/// ʂԂȂXe[ggsiINSERT, DELETE, UPDATE, CREATE TABLEȂǁj
void SqliteStatement::execute()
{
	int res = sqlite3_step(m_stmt);
	if( SQLITE_ROW != res &&
		SQLITE_DONE != res )
		checkResult(res);
}

/// Xe[ggsAʂ擾邽߂̃J[\Ԃ
SqliteCursor * SqliteStatement::executeQuery()
{
	return (new SqliteCursor(m_stmt, m_db));
}

/// Prepared Statement̃ChJ[hɁAۂ̒loCh
void SqliteStatement::setParam(int index, int val)
{
	int res = sqlite3_bind_int(m_stmt, index, val);
	checkResult(res);
}

/// Prepared Statement̃ChJ[hɁAۂ̒loCh
void SqliteStatement::setParam(int index, __int64 val)
{
	int res = sqlite3_bind_int64(m_stmt, index, val);
	checkResult(res);
}

/// Prepared Statement̃ChJ[hɁAۂ̒loCh
void SqliteStatement::setParam(int index, double val)
{
	int res = sqlite3_bind_double(m_stmt, index, val);
	checkResult(res);
}

/// Prepared Statement̃ChJ[hɁAۂ̒loCh
void SqliteStatement::setParam(int index, const _TCHAR* val)
{
	SqliteStringW str = SqliteString(val).toUTF16();
	int res = sqlite3_bind_text16(
		m_stmt, index, str.c_str(),
		(int)str.length()*sizeof(wchar_t),
		SQLITE_TRANSIENT);

	checkResult(res);
}

/// Prepared Statement̃ChJ[hɁAۂ̒loCh
void SqliteStatement::setParam(int index, const void* val, int nBytes, bool makeCopy)
{
	int res = sqlite3_bind_blob(m_stmt, index, val, nBytes,
		makeCopy ? SQLITE_TRANSIENT : SQLITE_STATIC);

	checkResult(res);
}

void SqliteStatement::checkResult(int rescode)
{
	if(SQLITE_OK != rescode)
	{
		SqliteStringW errMsg = (const wchar_t*)sqlite3_errmsg16(m_db);
		throw QueryException(errMsg.toANSI().c_str(), rescode);
	}
}


}; //end of namespace sqlite3plus

