/**************************************************************
*
*  Jul. 8, 2004
*
*  The author disclaims copyright to this source code.
*  In place of a legal notice, here is a blessing:
*
*    May you do good and not evil.
*    May you find forgiveness for yourself and forgive others.
*    May you share freely, never taking more than you give.
*
**************************************************************/
#ifndef SqliteString_H
#define SqliteString_H

#include <windows.h>
#include <tchar.h>
#include <string>

namespace sqlite3plus {


/// ANSIUNICODE(UTF16, UTF8)𑊌݂ɕϊ@\NX
/**
 * @param charT charwchar_tłȂ΂ȂȂB
 */
template<typename charT>
class SqliteStringT : public std::basic_string<charT>
{
public:
	SqliteStringT(void)
		: std::basic_string<charT>()
	{}

	SqliteStringT(const charT* initStr)
		: std::basic_string<charT>(initStr)
	{}

	SqliteStringT(const std::basic_string<charT>& initStr)
		: std::basic_string<charT>(initStr)
	{}

	SqliteStringT(const SqliteStringT<charT>& initStr)
		: std::basic_string<charT>(initStr)
	{}

	virtual ~SqliteStringT(void)
	{}

	const SqliteStringT<charT>& operator=(const charT* str)
	{
		assign(str);
		return *this;
	}

	const SqliteStringT<charT>& operator=(const std::basic_string<charT>& str)
	{
		assign(str);
		return *this;
	}

	const SqliteStringT<charT>& operator=(const SqliteStringT<charT>& str)
	{
		assign(str);
		return *this;
	}

	/// UTF8ɕϊ
	SqliteStringT<char>			toUTF8();
	/// ANSIɕϊiJ[ˑj
	SqliteStringT<char>			toANSI();
	/// UTF16ɕϊioCgI[_[̓vbgtH[ˑAIntelnCPUł̓gGfBA=UNICODELITTLEj
	SqliteStringT<wchar_t>		toUTF16();

	/// UTF8񂩂ASqliteStringTIuWFNg쐬
	static SqliteStringT<charT>	fromUTF8(const char* strUTF8);
	/// ANSI񂩂ASqliteStringTIuWFNg쐬
	static SqliteStringT<charT>	fromANSI(const char* strANSI);
	/// UTF16񂩂ASqliteStringTIuWFNg쐬
	static SqliteStringT<charT>	fromUTF16(const wchar_t* strUTF16);

protected:
	static int calcMbsBufferSize(const wchar_t* wideStr, bool utf8)
	{
		if(wideStr)
		{
			return WideCharToMultiByte(
				utf8 ? CP_UTF8 : CP_ACP, 0, wideStr, -1, NULL, 0, NULL, NULL);
		}
		else
		{
			return 0;
		}
	}

	static int calcWideBufferSize(const char* mbStr, bool utf8)
	{
		if(mbStr)
		{
			return MultiByteToWideChar(
				utf8 ? CP_UTF8 : CP_ACP, 0, mbStr, -1, NULL, 0);
		}
		else
		{
			return 0;
		}
	}

	static int wideToMbs(const wchar_t* wideStr, char* mbsBuf, int bufSize, bool utf8)
	{
		if(wideStr == NULL || mbsBuf == NULL)
		{
			return 0;
		}
		else
		{
			return WideCharToMultiByte(
				utf8 ? CP_UTF8 : CP_ACP, 0, wideStr, -1, mbsBuf, bufSize, NULL, NULL);
		}
	}

	static int mbsToWide(const char* mbStr, wchar_t* wideBuf, int bufSize, bool utf8)
	{
		if(mbStr == NULL || wideBuf == NULL)
		{
			return 0;
		}
		else
		{
			return MultiByteToWideChar(
				utf8 ? CP_UTF8 : CP_ACP, 0, mbStr, -1, wideBuf, bufSize);
		}
	}
};


// inline specializations

// char specialization
template <>
inline SqliteStringT<char> SqliteStringT<char>::toANSI()
{
	return *this;
}
// wchar_t specialization
template <>
inline SqliteStringT<char> SqliteStringT<wchar_t>::toANSI()
{
	SqliteStringT<char> ret;

	// calculate destination buffer size
	int nBytes = calcMbsBufferSize(c_str(), false);
	CHAR* mbsBuf = (CHAR*)malloc(sizeof(CHAR)*(nBytes+1)); // including trailing NULL char

	// convert to ANSI string
	wideToMbs(c_str(), mbsBuf, nBytes, false);
	mbsBuf[nBytes] = '\0';
	ret = mbsBuf;

	free(mbsBuf);

	return ret;
}


// char specialization
template <>
inline SqliteStringT<wchar_t> SqliteStringT<char>::toUTF16()
{
	SqliteStringT<wchar_t> ret;

	int nChars = calcWideBufferSize(c_str(), false);
	WCHAR* buf = (WCHAR*)malloc(sizeof(WCHAR)*(nChars+1));
	mbsToWide(c_str(), buf, nChars, false);
	buf[nChars] = L'\0';
	ret = buf;
	free(buf);

	return ret;
}
// wchar_t specialization
template <>
inline SqliteStringT<wchar_t> SqliteStringT<wchar_t>::toUTF16()
{
	return *this;
}


// char specialization
template <>
inline SqliteStringT<char> SqliteStringT<char>::toUTF8()
{
	SqliteStringT<char> ret;

	// first, convert to unicode string
	int nWideChars = calcWideBufferSize(c_str(), false);
	WCHAR* wideBuf = (WCHAR*)malloc(sizeof(WCHAR)*(nWideChars+1));

	// convert to Unicode
	mbsToWide(c_str(), wideBuf, nWideChars, false);
	wideBuf[nWideChars] = L'\0';

	// convert to UTF8
	int nBytes = calcMbsBufferSize(wideBuf, true);
	CHAR* utfBuf = (CHAR*)malloc(sizeof(CHAR)*(nBytes+1));
	wideToMbs(wideBuf, utfBuf, nBytes, true);
	utfBuf[nBytes] = '\0';

	ret = utfBuf;

	free(utfBuf);
	free(wideBuf);

	return ret;
}
// wchar_t specialization
template <>
inline SqliteStringT<char> SqliteStringT<wchar_t>::toUTF8()
{
	SqliteStringT<char> ret;

	// calculate destination buffer size
	int nBytes = calcMbsBufferSize(c_str(), true);
	CHAR* utfBuf = (CHAR*)malloc(sizeof(CHAR)*(nBytes+1)); // including trailing NULL char

	// convert to UTF8
	wideToMbs(c_str(), utfBuf, nBytes, true);
	utfBuf[nBytes] = '\0';
	ret = utfBuf;

	free(utfBuf);

	return ret;
}


// char specialization
template <>
inline SqliteStringT<char> SqliteStringT<char>::fromUTF8(const char* strUTF8)
{
	SqliteStringT<char> ret;

	// convert to Unicode
	int nWideChars = calcWideBufferSize(strUTF8, true);
	WCHAR* wideBuf = (WCHAR*)malloc(sizeof(WCHAR)*(nWideChars+1));
	mbsToWide(strUTF8, wideBuf, nWideChars, true);

	// convert to ANSI
	int nBytes = calcMbsBufferSize(wideBuf, false);
	CHAR* mbsBuf = (CHAR*)malloc(sizeof(CHAR)*(nBytes+1));
	wideToMbs(wideBuf, mbsBuf, nBytes, false);
	mbsBuf[nBytes] = '\0';

	ret = mbsBuf;

	free(mbsBuf);
	free(wideBuf);

	return ret;
}
// wchar_t specialization
template <>
inline SqliteStringT<wchar_t> SqliteStringT<wchar_t>::fromUTF8(const char* strUTF8)
{
	SqliteStringT<wchar_t> ret;

	// convert to Unicode
	int nWideChars = calcWideBufferSize(strUTF8, true);
	WCHAR* wideBuf = (WCHAR*)malloc(sizeof(WCHAR)*(nWideChars+1));
	mbsToWide(strUTF8, wideBuf, nWideChars, true);
	wideBuf[nWideChars] = L'\0';

	ret = wideBuf;

	free(wideBuf);

	return ret;
}


// char specialization
template <>
inline SqliteStringT<char> SqliteStringT<char>::fromANSI(const char* strANSI)
{
	return SqliteStringT<char>(strANSI);
}
// wchar_t specialization
template <>
inline SqliteStringT<wchar_t> SqliteStringT<wchar_t>::fromANSI(const char* strANSI)
{
	return SqliteStringT<char>(strANSI).toUTF16();
}


// char specialization
template <>
inline SqliteStringT<char> SqliteStringT<char>::fromUTF16(const wchar_t* strUTF16)
{
	return SqliteStringT<wchar_t>(strUTF16).toANSI();
}
// wchar_t specialization
template <>
inline SqliteStringT<wchar_t> SqliteStringT<wchar_t>::fromUTF16(const wchar_t* strUTF16)
{
	return SqliteStringT<wchar_t>(strUTF16);
}


// typedef
/// }`oCgLN^Zbg̕iANSIyUTF8jێNX
typedef SqliteStringT<char>		SqliteStringA;
/// ChLN^Zbg̕iUTF16jێNX
typedef SqliteStringT<wchar_t>	SqliteStringW;
/// rhɂASqliteStirngA, SqliteStringŴǂ炩ƓȂ
typedef SqliteStringT<_TCHAR>	SqliteString;


}; //end of namespace sqlite3plus

#endif SqliteString_H
