
# to TDF
use strict;

# $0  TDS-file-dir  index-file-name  input-file




use utf8;
my $ENCODE_NAME = ':encoding(shift_jis)';
binmode(STDIN,  $ENCODE_NAME);
binmode(STDOUT, $ENCODE_NAME);
binmode(STDERR, $ENCODE_NAME);

# TODO: 外部から指定できるようにしよう
my @AllPluginNames = qw(util_system util_generic util_graph util_sqlite util_io util_iconv);
my $OtherDocumentLinks = join "\n", map { qq(DMI link rel="Chapter" href="../${_}/index.html" title="プラグイン ${_}"); } (@AllPluginNames);

# TODO: グローバル変数を減らそう
my $CurTag = '';	# 処理中のタグ名
my $CurExt = '';	# 処理中のタグ引数
my @Texts = ();
my $DocStruct = {};
my @MemberList = ();
my $CurName = '';	# 処理中のメンバ名
my $Headings = {};	# メソッド名と見出し番号の関連付け
my $HeadingCounter = 2;
my $PlugInName = '';
my $DMI_CONST = << "END";
DMI link rel="stylesheet" type="text/css" href="../../style/tp_utils.css"
DMI link rel="Start" href="../../index.html" title="トップ"
DMI link rel="Help" href="http://devdoc.kikyou.info/tvp/docs/tjs2doc/contents/" title="TJS2オンラインマニュアル"
DMI link rel="Help" href="http://devdoc.kikyou.info/tvp/docs/kr2doc/contents/" title="吉里吉里２オンラインマニュアル"
END


my $TDFDir = shift(@ARGV) || '../doc/tdf';
my $IndexFileName = shift(@ARGV) || 'index.tdf';
my $InFile = shift(@ARGV) || 'ファイル名が指定されていません。';

my $TitleName = '名無し';
my $FileContent = {};

my $IsProperty = 0;
my %MemberTypes = ();	# 0:func, 1:prop, 2:constructor
my %MemberSynopsys = ();

# 外部マニュアル
my %ExternalRoot = (
	'!TVP' => 'http://devdoc.kikyou.info/tvp/docs/kr2doc/contents',
	'!TJS' => 'http://devdoc.kikyou.info/tvp/docs/tjs2doc/contents',
);

my %HList = (
'TITLE'       => ['TITLE', ''],
'DECLARATION' => ['TITLE', ''],
#'ABSTRACT'    => ['P', '/P'],
'ABSTRACT'    => ['SECTION 概要', '/SECTION'],
'SYNOPSYS'    => ['P', '/P'],
'EXCEPTION'   => ['SECTION 例外', '/SECTION'],
'DESCRIPTION' => ['SECTION 詳細', '/SECTION'],
'EXAMPLE'     => ['SECTION 例', '/SECTION'],
'ARGUMENT'    => ['SECTION 引数', '/SECTION'],
'RETURN'      => ['SECTION 戻り値', '/SECTION'],
'SEE'         => ['SECTION 関連項目', '/SECTION'],
'BUGS'        => ['SECTION バグ', '/SECTION'],
'GETTER'      => ['SECTION 値の取得', '/SECTION'],
'SETTER'      => ['SECTION 値の設定', '/SECTION'],
'EX'          => ['SECTION ', '/SECTION'],
);

my %ArgTypes = (
'Integer'  => '整数型',
'Real'     => '実数型',
'Object'   => 'オブジェクト型',
'Function' => '関数オブジェクト型',
'String'   => '文字列型',
'Octet'    => 'オクテット列型',
'Any'      => '任意の型、任意の数',
'Array'    => '配列型',
'Dictionary' => '辞書配列型',
'Layer' => 'レイヤオブジェクト型',
);



print STDERR "入力ファイル: $InFile\n";
print STDERR "TDF出力先   : $TDFDir\n";
print STDERR "索引ファイル: $IndexFileName\n";


open (IF, "$InFile") or die("$InFile: $!\n");
binmode(IF, $ENCODE_NAME);

while (<IF>) {
	next if (/^#/);
	s/\n//;
	# タグ
	if (/^[A-Z]/s) {
		my $x = $_;
		out($CurTag, $CurExt, \@Texts);
		($CurTag, $CurExt) = split(/\s+/, $x, 2);
		@Texts = ();
	} elsif (/^\t(.*)/) {
		my $t = $1;
		if ($t =~ /^\s*$/) {
			$t .= '<<CR>>';
		}
		push (@Texts, $t);
	}
}
close(IF);


out($CurTag, $CurExt, \@Texts);

# 各ファイルに出力
while (my ($k, $v) = (each %{$FileContent})) {
	next if ($k eq '_');
	my $file = "${TDFDir}/" . getFileNameBase($k) . '.tdf';
	open (F, ">$file") or die("$file: $!");
	binmode(F, $ENCODE_NAME);
	print F $v;
	close (F);
}

outIndex();


sub getFileNameBase {
	my ($f) = @_;
	return "f_${f}";
}


sub outIndex {
	open (F, ">$TDFDir/$IndexFileName") or die("$TDFDir/$IndexFileName: $!");
	binmode(F, $ENCODE_NAME);

	if ($PlugInName) {
print F << "END";
$FileContent->{'_'}

SECTION メンバ一覧
P
このプラグインをリンクすると、下記のメンバが追加されます。
/P
UL
END
	} else {
print F << "END";
$FileContent->{'_'}

SECTION メンバ一覧
P
下記のメンバが追加されています。
/P
UL
END
	}

	my @memberTypeStrs = ('関数', 'プロパティ', 'コンストラクタ');
	foreach (@MemberList) {
		my $memberTypeStr = $memberTypeStrs[$MemberTypes{$_}];
		my $memberSyn = $MemberSynopsys{$_};
		my $fn = getFileNameBase($_) . '.html';
		print F << "EOF";
LI+
LINK $fn $_
 （$memberTypeStr：$memberSyn）
/LI
EOF
	}
print F << "END";
/UL
/SECTION
END
	close (F);
}



sub out {
	my ($tag, $tagext, $texts) = @_;
	my $rv = '';
	my $text = join("\n", @$texts);
	my $prp = 0;

	my $btag = $HList{$tag}->[0];
	my $etag = $HList{$tag}->[1];

	if ($tag eq 'PLUGIN') {
		$PlugInName = $text;
#		$TDFDir .= "/$PlugInName";
		#print STDERR $PlugInName;

	} elsif ($tag eq 'TITLE') {
		$rv = "$btag $text\n\n";
		$TitleName = $text;
		print $rv;

	} elsif ($tag eq 'SEE') {
		$rv = "$btag\nUL\n";
		my @members = split(/\t*\n\t*/, $text);
		foreach (@members) {
			my $link = createLink("$_ -");
			$rv .= "LI* ${link}\n" if ($link);
		}
		$rv .= "$etag\n";

	} elsif ($tag eq 'EXCEPTION') {
		$rv = "$btag\nUL\n";
		my @excs = split(/\t*\n\t*/, $text);
		foreach (@excs) {
			my $t = $_;
			$t = asTDSIL($t);
			$rv .= "LI+\n${t}\n/LI\n";
		}
		$rv .= "$etag\n";

	} elsif ($tag eq 'DECLARATION') {
		my $static = $text =~ s/^\s*static\s+//;
		my $othercontext = $text =~ s/^\s*othercontext\s+//;

		my ($name, $args) = ('', '');

		if ($text =~ /(^.*)\((.*)\)/) {
			($name, $args) = ($1, $2);
			$IsProperty = 0;
			$MemberTypes{$name} = 0;
		} else {
			$text =~ /(^.*)/;
			$name = $1;
			$IsProperty = 1;
			$MemberTypes{$name} = 1;
		}

		$CurName = $name;
		my ($ctx, $member, $ctor) = ('global', $name, '');
		my $ctx_additional = '';
		my $global = 1;
		# with context (except global)
		if ($name =~ /\./) {
			($ctx, $member) = split(/\./, $name, 2);
			$ctor = ($member eq $ctx) ? 1 : '';
			# constructor
			if ($ctor) {
				$MemberTypes{$name} = 2;
				$ctx = "global";
				$ctor = "new $ctx.";
			# others
			} else {
				$ctx = "global.$ctx";
				$ctx_additional = ($static) ? 
					' (static)' :
					' インスタンス';
			}
			$global = 0;
		}

		# owner
		my $owner = ($ctor) ? "global.$member" : "$ctx$ctx_additional";

		# with arguments
		if ($args) {
			my @aa = split(/\s*,\s*/, $args);
			my @ar = ();
			foreach my $t (@aa) {
				my $id = getArgID($CurName, $t);
				push @ar, "LINK #${id} ${t}";
			}
			$args = "\n" . join("\n, \n", @ar) . "\n";
		}

		# how to call
		my $callcode;
		my $argCallStr = $IsProperty ? '' : "($args )";
		my $memberTypeStr = $IsProperty ? 'プロパティ' : 
			(($ctor) ? 'コンストラクタ' : '関数');

		if ($othercontext) {
			$callcode = "($ctx.$member incontextof\nVAR object\n)$argCallStr";
		} elsif ($static) {
			$callcode = "$ctx.$member$argCallStr";
		} elsif ($ctor) {
			$callcode = "$ctor$member$argCallStr";
		} elsif ($global) {
			$callcode = "$ctx.$member$argCallStr";
		} else {
			$callcode = "VAR instance\n.$member$argCallStr";
		}

		push @MemberList, $name;

		$rv = << "END";
$btag $name （<!--SYNOPSYS-->）

SECTION 文法
DL class='memberabstract'
DT メンバ型
DD $memberTypeStr
DT メンバ所有者
DD+
ICODE $owner
/DD
DT 呼び出し
DD+
ICODE+
$callcode
/ICODE
/DD
/DL
/SECTION
END

	} elsif ($tag eq 'ARGUMENT') {
		$text = getArgList($texts);
		$text = "DL class='arglist'\n$text\n/DL\n";
		$text = asTDSIL($text);
		$rv = "$btag\n$text\n$etag\n\n";

	} elsif ($tag eq 'SYNOPSYS') {
		$rv = << "END";
$btag
$text
$etag

END

		$DocStruct->{'DECLARATION'} =~ s/<!--SYNOPSYS-->/$text/;
		$MemberSynopsys{$CurName} = $text;

	} elsif (grep(/^$tag$/, qw(DESCRIPTION EXAMPLE ABSTRACT SETTER GETTER BUGS))) {
		$text = asTDS($text, 'P');
		$text = asTDSIL($text);
		$rv = << "END";
$btag
P
$text
/P
$etag

END

	} elsif ($tag eq 'EX') {
		$text = asTDS($text, 'P');
		$text = asTDSIL($text);
		$rv = << "END";
$btag $tagext
P
$text
/P
$etag

END

	} elsif ($tag eq 'RETURN') {
		$prp = 1;
		$text = asTDS($text, 'P');
		$text = asTDSIL($text);
		if ($IsProperty) {
			$rv = << "END";
$HList{'DECLARATION'}->[1]

END
		} else {
			$rv = << "END";
$btag
P
$text
/P
$etag
$HList{'DECLARATION'}->[1]

END
		}
	}

	$rv = removeEmptyPair($rv, 'P');
	$rv = removeEmptyPair($rv, 'DD');

	if ($tag eq 'EX') {
		$DocStruct->{$tag} .= $rv;
	} else {
		$DocStruct->{$tag} = $rv;
	}

	# メンバのページ
	if ($prp) {
		$FileContent->{$CurName} = << "END";
DMI link rel="Toc" href="index.html#h2." title="目次"
${DMI_CONST}
END

		if ($PlugInName) {
			$FileContent->{$CurName} .= << "END";
${OtherDocumentLinks}
END
		}

		$FileContent->{$CurName} .= << "END";
$DocStruct->{'DECLARATION'}
$DocStruct->{'DESCRIPTION'}
$DocStruct->{'ARGUMENT'}
$DocStruct->{'GETTER'}
$DocStruct->{'SETTER'}
$DocStruct->{'EXCEPTION'}
$DocStruct->{'RETURN'}
$DocStruct->{'EXAMPLE'}
$DocStruct->{'EX'}
$DocStruct->{'SEE'}
END
		$DocStruct = {};
		$Headings->{$CurName} = $HeadingCounter++;
	}

	# インデックス用
	if ($DocStruct->{'TITLE'}) {
		if ($PlugInName) {
			$FileContent->{'_'} = << "END";
${DMI_CONST}
${OtherDocumentLinks}
TITLE 吉里吉里プラグイン $PlugInName：$TitleName
$DocStruct->{'ABSTRACT'}
END
		} else {
			$FileContent->{'_'} = << "END";
${DMI_CONST}
TITLE $TitleName
$DocStruct->{'ABSTRACT'}
END
		}
	}
}

# TDSブロック
# <<TDS～>> ベタTDS
sub asTDS {
	my ($text, $t) = @_;
	my $te = '';
	my $r = '';

	my @blocks = split(/((?:\n|^)<<TDS.*?\n>>)/s, $text);

	$te = "/$t" if ($t);
	foreach my $b (@blocks) {
		if ($b =~ /^\n?<<TDS/) {
			my $withIL = defined($1);
			$b =~ s/\n<<TDS/\n$te/;
			$b =~ s/^<<TDS/$te/;
			$b =~ s/\n>>/\n$t\n/;
			$b = replaceCR($b, '');
#print STDERR "TDS $t\n";
		} else {
			$b = replaceCR($b, $t);
#print STDERR "NRM $t\n";
		}
		$r .= $b;
	}
	return $r;
}

# TDSコマンドとlinkコマンド
sub asTDSIL {
	my ($text) = @_;

# @{link method arg disp}
	while ($text =~ /(.?)\@\{link\s+(.*?)\}(.?)/sg) {
		my $a = $2;
		my $pcr = (defined($1)) ? $1 : '';
		my $xcr = (defined($3)) ? $3 : '';

		my $link = createLink($a);
		if ($link) {
			$pcr = "$pcr\n" if ($pcr ne '' && $pcr ne "\n");
			$xcr = "\n$xcr" if ($xcr ne '' && $xcr ne "\n");
			$text =~ s/.?\@\{link\s+(.*?)\}.?/${pcr}LINK ${link}${xcr}/s;
		}
	}

# @{COMMAND params}
# ->
# COMMAND params
	while ($text =~ /(.?)\@\{(.*?)\}(.?)/sg) {
		my $a = $2;
		my $pcr = (defined($1)) ? $1 : '';
		my $xcr = (defined($3)) ? $3 : '';

		$pcr = "$pcr\n" if ($pcr ne '' && $pcr ne "\n");
		$xcr = "\n$xcr" if ($xcr ne '' && $xcr ne "\n");
		$text =~ s/.?\@\{(.*?)\}.?/${pcr}${a}${xcr}/s;
	}

	return $text;
}


# リンク作成
# @{link MEMBER ARG DISP}
sub createLink {
	my $rv = '';
	my ($a) = @_;
	my @as = split(/\s+/, $a, 3);

	# ARGがない場合は '-' が指定されたとみなす。
	if (scalar(@as) == 1) {
		$as[1] = '-';
	}

	if (scalar(@as) >= 2) {
		my $ac;
		my $exfile = '';
		my $methodOmit = 0;
		my $argOmit = 0;
		my $pluginName = '';

		# オンラインマニュアル
		my $online = 0;

		# 外部ファイル参照
		if ($as[0] =~ /^.*::/) {
			$as[0] =~ s/^(.*):://;
			$exfile = $1;
			$exfile =~ s/^\s*//;
			# 外部マニュアルへの参照
			# !TVP, !TJS
			if (exists($ExternalRoot{$exfile})) {
				$online = $exfile;
				$exfile = $ExternalRoot{$exfile};

			# 他のプラグインへの参照
			# !util_xxx
			} elsif ($exfile =~ /^!(\S+)/) {
				$pluginName = $1;
				$exfile = '';
			}
		}

		# クラス＋メソッド省略時
		if ($as[0] eq '-') {
			$as[0] = $CurName;
			$methodOmit = 1;
		}
		# 引数名省略時
		if ($as[1] eq '-') {
			$ac = '';
			$argOmit = 1;
		# 引数名あり
		} else {
			$ac = getArgID($as[0], $as[1]);
		}
		my $ds = $as[2];
		# 表示文字列省略時
		if ($ds eq '') {
			$ds = $argOmit ? $as[0] : $as[1];
		}
		unless ($exfile) {
			$exfile = getFileNameBase($as[0]) . '.html';
		}
		my $url = '';

		# 命名規則: URL = $TVPDocRoot/f_CLASS_METHOD.html
		if ($online eq '!TVP') {
			my $a = $as[0];
			$a =~ s/\./_/;
			$url = "${exfile}/f_${a}.html";

		# 命名規則: URL = $TJSDocRoot/dictionary.html
		} elsif ($online eq '!TJS') {
			my $a = lc($as[0]);
			# Math.RandomGenerator対策
			if ($a =~ /math\.randomgenerator/) {
				$a = 'randomgenerator';
			} else {
				$a =~ s/\..*//;
			}
			$url = "${exfile}/${a}.html";

		# 命名規則: URL = #ARG
		} elsif ($methodOmit) {
			$url = "#${ac}";

		# 命名規則: URL = FILEBASE#ARG
		} else {
			# 他のプラグイン
			$url = "../$pluginName/" if ($pluginName);
			$url .= ${exfile};
			$url .= "#${ac}" if ($ac);
		}

		$rv = "${url} ${ds}";
	}

	return $rv;
}



sub removeEmptyPair {
	my ($text, $t) = @_;
	$text =~ s|\n$t\n+\/$t\n|\n|sg;
	return $text;
}

sub replaceCR {
	my ($text, $t) = @_;

	if ($t) {
		my $te = "/$t";
		$text =~ s/^\s*//;
		$text =~ s|\n<<CR>>|\n$te\n$t|g;
		$text =~ s|^<<CR>>|$te\n$t|g;
	} else {
		$text =~ s/(?:\n|^)<<CR>>/\n/g;
	}
	return $text;
}


sub getArgID {
	my ($name, $arg) = @_;
	$arg = 'optional-arguments' if ($arg eq '*');
	return "ARG.$arg";
}

sub getMethodID {
	my ($name) = @_;
	return "$name";
}



sub getArgList {
	my ($texts) = @_;
	my $rv = '';
	my $dd = 0;
	my $hd = 0; # here doc

	foreach my $e (@$texts) {
		if ($e =~ /^\t*<<TDS$/) {
			$hd = 1;
			next;
		}

		if ($e =~ /^\t*>>$/) {
			$hd = 0;
			next;
		}

		if ($hd) {
			$e =~ s/^\t*//;
			$rv .= "$e\n";

		} elsif ($e =~ /^\t/) {
			unless ($dd) {
				$rv .= "DD+\n";
				$dd = 1;
			}
			$e =~ s/^\s*//;
#			$e = replaceCR($e, 'DD');
			$e = asTDS($e, 'DD');
			$rv .= "$e\n";

		} else {
			if ($dd) {
				$rv .= "/DD\n";
				$dd = 0;
			}
			my @es = split(/\s+/, $e, 3);
			# 引数タイプ
			my @ats = ();
			foreach my $e (split(/\|/, $es[1])) {
				if (exists($ArgTypes{$e})) {
					push @ats, qq(\nVERB <dfn title="$ArgTypes{$e}">$e</dfn>);
				} else {
					push @ats, " $e";
				}
			}
			$es[1] = ('(' . join(' | ', @ats) . ' )');
			# デフォルト値
			if ($es[2]) {
				if ($es[2] =~ /^\[(.*)\]$/) {
					$es[2] = "[default: $1]";
				}
			}
			$e = join(' ', @es);
			my $id = getArgID($CurName, $es[0]);
			$rv .= "DT+ id='$id'\n$e\n/DT\n";
		}
	}

	if ($dd) {
		$rv .= "/DD\n";
		$dd = 0;
	}

	return $rv;
}

