
#
# DLLプラグイン用
#

use strict;
use utf8;
use File::Basename;
use lib (dirname($0));
use TVPPlugin;

binmode(STDIN,  $ENCODE_NAME);
binmode(STDOUT, $ENCODE_NAME);


my $CLASS_DEF = '';
my $EXTERN_VARS = '';
my %buf = ();
my $USE_DH = $ARGV[0] ? 1 : 0;	# データホルダ使用？

my $DefineID = genDefineID();

shift @ARGV;	# 残りはinclude
my $EXTINC = join('', map { qq|#include "$_"\n| } @ARGV);


my $temp_classdef = << 'END';
// Function : ###FUNC_NAME###
class ###CLASS_NAME### : public tTJSDispatch {
public:
	// FuncCall decl.
	tjs_error TJS_INTF_METHOD FuncCall(
		tjs_uint32 flag, const tjs_char * membername, tjs_uint32 *hint,
		tTJSVariant *result, tjs_int numparams, tTJSVariant **param, iTJSDispatch2 *objthis);

###CLASS_IMPL###
};

END


my $TypeFunc = 1;

my ($cppclass, $f, $impl);

my @Data = ();


while (<STDIN>) {
	chomp;
	push @Data, $_;
}

foreach (@Data) {
	# 続き
	if ($_ =~ /^::(\S+)/s) {
		$cppclass = getCppClassName($1);
		$impl = '';
		while (<STDIN>) {
			last if ($_ =~ /^;;/s);
			$impl .= $_;
		}
		$buf{$cppclass} =~ s/###CLASS_IMPL###/${impl}/gs;

	} else {
		chomp;
		$_  =~ s/^\s+//;
		next if (!$_ || $_ =~ /^#/);
		$f = $_;
		my $ptype = $TypeFunc;
		my $pvalue = '(tTVInteger)0';

		# クラス
		if ($f =~ s/\@class//) {
			next;
		}

		# スタティック
		if ($f =~ s/\@static//) {
			# no action
		}

		# プロパティ： 
		if ($f =~ s/:([rw]+)//) {
			next;
		}

		if ($f && ($ptype == $TypeFunc)) {
			$cppclass = getCppClassName($f);
			my $t = $temp_classdef;
			$t =~ s/###CLASS_NAME###/$cppclass/gs;
			$t =~ s/###FUNC_NAME###/$f/gs;
			$t =~ s/###PROP_VALUE###/$pvalue/gs;
			$buf{$cppclass} = $t;
		}
	}
}


while (my ($k, $v) = (each %buf)) {
	$v =~ s/###CLASS_IMPL###//gs;

	$CLASS_DEF .= $v;
}


my $temp_base = << "END";
//---------------------------------------------------------------------------
// Generated by $0
//---------------------------------------------------------------------------
#ifndef ${DefineID}
#define ${DefineID}

END

# use holder ?
if ($USE_DH) {
	$temp_base .= << 'END';
#include "TableHolder.h"

extern CTableHolder * TableHolder;

END
}

# リンク・アンリンクイベント
foreach my $func qw(V2Link V2Unlink) {
	foreach my $ev qw(Enter Exit) {
		$temp_base .= << "END";
void TPSet${func}${ev}Callback(void (*callback)(void));
END
	}
}


$temp_base .= << "END";

// 実装せよ
void TPSetLinkEventCallback(void);


$EXTINC
#include "TVPPlugIn.h"

${CLASS_DEF}
${EXTERN_VARS}
#endif ${DefineID}
//---------------------------------------------------------------------------
END

print $temp_base;

sub genDefineID {
	my $a = $0;
	$a =~ s|\\|/|g;
	$a =~ s|.*/||;
	$a =~ tr/./_/;
	$a = uc($a);
	my $rv .= "_TVP_PLUGIN_AUTOGEN_${$}_" . time() . "_${a}_H_";
	return $rv;
}
