
# オブジェクトインターフェース
use strict;

use utf8;
my $ENCODE_NAME = ':encoding(shift_jis)';
binmode(STDIN,  $ENCODE_NAME);
binmode(STDOUT, $ENCODE_NAME);

exit if ($#ARGV < 1);
my ($OutSrcH, $OutSrcC) = @ARGV;

my $HeaderID = '_TVP_PLUGIN_TJSDISPATCH2_INTERFACE_EX_AUTO_H_';

my $PropGets = [];
my $PropGetByNums = [];


my $Delim = '//' . ('-' x 78);

my $PropGetTemplate = << 'EOH';
###DELIMITER###
// property get ###CLASS_NAME###.###PROP_NAME###
###DELIMITER###
/*TP*/const ###PROP_TYPE### ###CLASS_NAME###_PropGet_###PROP_NAME###(iTJSDispatch2 * obj, const tjs_char * memberName)
{
	static tTJSVariant workVal;
	static tjs_uint32 hint = 0;
	if (TJS_FAILED(obj->PropGet(TJS_MEMBERMUSTEXIST, TJS_W("###PROP_NAME###"), &hint, &workVal, obj))) {
		TP_ThrowMes(TJSEM_E_PROPGET_FAIL, memberName, TJS_W("###CLASS_NAME###") TJS_W(".") TJS_W("###PROP_NAME###"));
	}
	return ###TYPE_CAST###workVal###TYPE_CAST2###;
}

###DELIMITER###
// property get ###CLASS_NAME###.###PROP_NAME### (anonymous)
###DELIMITER###
/*TP*/const ###PROP_TYPE### ###CLASS_NAME###_PropGet_###PROP_NAME###(iTJSDispatch2 * obj)
{
	return ###CLASS_NAME###_PropGet_###PROP_NAME###(obj, TJSEM_ANON_FUNC);
}

EOH


my $PropGetByNumTemplate = << 'EOH';
###DELIMITER###
// property get (by num) ###CLASS_NAME###[n] as ###PROP_TYPE###
###DELIMITER###
/*TP*/const ###PROP_TYPE### ###CLASS_NAME###_PropGetByNumAs###TYPE_ID###(iTJSDispatch2 * obj, tjs_int num, const tjs_char * memberName)
{
	static tTJSVariant workVal;
	if (TJS_FAILED(obj->PropGetByNum(TJS_MEMBERMUSTEXIST, num, &workVal, obj))) {
		TP_ThrowMes(TJSEM_E_PROPGETBYNUM_FAIL, memberName, TJS_W("###CLASS_NAME###"), (ttstr)num);
	}
	return ###TYPE_CAST###workVal###TYPE_CAST2###;
}

###DELIMITER###
// property get (by num) ###CLASS_NAME###[n] as ###PROP_TYPE###
###DELIMITER###
/*TP*/const ###PROP_TYPE### ###CLASS_NAME###_PropGetByNumAs###TYPE_ID###(iTJSDispatch2 * obj, tjs_int num)
{
	return ###CLASS_NAME###_PropGetByNumAs###TYPE_ID###(obj, num, TJSEM_ANON_FUNC);
}
EOH


my $HeadSrc = << "END";
//----------------------------
// Generated by $0
//----------------------------

END


my $OutC = << "EOH";
$HeadSrc

#include "TVPPlugIn.h"
#include "$OutSrcH"

EOH

my $OutH = << "EOH";
$HeadSrc
#ifndef $HeaderID
#define $HeaderID

#include <windows.h>
#include "tp_stub.h"

EOH



while (<STDIN>) {
	chomp;
	$_  =~ s/^\s+//;
	next if (!$_ || $_ =~ /^#/);
	my ($a, $f) = split(/\s*,\s*/, $_, 2);
	if ($a && $f) {
		if ($a =~ s/^!//) {
			my ($class, $asname) = split(/\s+/, $f);
			push @$PropGetByNums, [$a, $class, $asname];
		} else {
			my ($class, $asname) = split(/\./, $f);
			push @$PropGets, [$a, $class, $asname];
		}

	}
}


foreach my $ea (@$PropGets) {
	my ($type, $class, $name) = @$ea;
	my $t = $PropGetTemplate;
	my $cast = '';
	my $cast2 = '';
	my $op = '';
	# dispatch
	if ($type =~ /iTJSDispatch2\s*\*$/) {
		$cast2 = "->AsObjectNoAddRef()";
	} elsif ($type =~ /tTJSVariantOctet\s*\*$/) {
		$cast2 = ".AsOctetNoAddRef()";
	} elsif ($type =~ /.+\*$/) {
		$cast = "($type)(tTVInteger)";
	} elsif ($type eq 'tjs_real') {
		$cast = "(tTVReal)";
	} elsif ($type ne 'tjs_int') {
		$cast = "($type)(tjs_int)";
	}

	$t =~ s/###PROP_TYPE###/$type/g;
	$t =~ s/###CLASS_NAME###/$class/g;
	$t =~ s/###PROP_NAME###/$name/g;
	$t =~ s/###TYPE_CAST###/$cast/g;
	$t =~ s/###TYPE_CAST2###/$cast2/g;
	$t =~ s/###DELIMITER###/$Delim/g;

	while ($t =~ m|/\*TP\*/(.*?)\n|sg) {
		$OutH .= "$1;\n";
	}

	$t =~ s|/\*TP\*/||g;
	$OutC .= "$t\n";
}

foreach my $ea (@$PropGetByNums) {
	my ($type, $class, $id) = @$ea;
	my $t = $PropGetByNumTemplate;
	my $cast = '';
	my $cast2 = '';
	# dispatch
	if ($type =~ /iTJSDispatch2\s*\*$/) {
		$cast2 = ".AsObjectNoAddRef()";
	} elsif ($type =~ /tTJSVariantOctet\s*\*$/) {
		$cast2 = ".AsOctetNoAddRef()";
	} elsif ($type =~ /.+\*$/) {
		$cast = "($type)(tjs_int64)";
	} elsif ($type eq 'tjs_real') {
		$cast = "(tTVReal)";
	} elsif ($type ne 'tjs_int') {
		$cast = "($type)(tjs_int)";
	}

	# reference
	if ($type =~ /.+\&$/) { $cast = "($type)" };

	$t =~ s/###PROP_TYPE###/$type/g;
	$t =~ s/###CLASS_NAME###/$class/g;
	$t =~ s/###TYPE_ID###/$id/g;
	$t =~ s/###TYPE_CAST###/$cast/g;
	$t =~ s/###TYPE_CAST2###/$cast2/g;
	$t =~ s/###DELIMITER###/$Delim/g;

	while ($t =~ m|/\*TP\*/(.*?)\n|sg) {
		$OutH .= "$1;\n";
	}

	$t =~ s|/\*TP\*/||g;
	$OutC .= "$t\n";
}

$OutH .= << "EOH";

#endif $HeaderID
EOH

open FC, ">$OutSrcC";
open FH, ">$OutSrcH";

binmode(FC,  $ENCODE_NAME);
binmode(FH,  $ENCODE_NAME);

print FC $OutC;
print FH $OutH;

close FC;
close FH;

exit;
