
#
# DLLプラグイン用
#

use strict;
use utf8;
use File::Basename;
use lib (dirname($0));
use TVPPlugin;

binmode(STDIN,  $ENCODE_NAME);
binmode(STDOUT, $ENCODE_NAME);


my $REGISTER = '';
my $REGISTER_PROP = '';
my $RELEASE = '';
my $RELEASE_PROP = '';
my $VARS = '';
my $FUNCS = '';
my $HNAME_BASE = $ARGV[0] || 'NoName';
my $HNAME = "$HNAME_BASE.h";
my $USE_DH = $ARGV[1] ? 1 : 0;	# データホルダ使用？
my %ICO_CLASS = ();

my $PluginOption1 = "${HNAME_BASE}-verbose";


my @RegisterList = ();
my @ReleaseList = ();


# グローバル関数
my $temp_register = << 'END';
	//-----------------------------------------------------------------------
	TP_ADD_LOG_REG_MEMBER(global.###FUNC_NAME###);
	###VAR_NAME### = new ###CLASS_NAME###();
	val = tTJSVariant(###VAR_NAME###);
	###VAR_NAME###->Release();
	global->PropSet(TJS_MEMBERENSURE ###CALL_FLAG###, TJS_W("###FUNC_NAME###"), NULL, &val, global);
	//-----------------------------------------------------------------------
END

# メソッド
my $temp_register_wc = << 'END';
	//-----------------------------------------------------------------------
	TP_ADD_LOG_REG_MEMBER(global.###ICO_CLASS###.###FUNC_NAME###);
	###VAR_NAME### = new ###CLASS_NAME###();
	val = tTJSVariant(###VAR_NAME###);
	###VAR_NAME###->Release();
	classdef_###ICO_CLASS###->PropSet(TJS_MEMBERENSURE ###CALL_FLAG###, TJS_W("###FUNC_NAME###"), NULL, &val, classdef_###ICO_CLASS###);
	//-----------------------------------------------------------------------
END

# クラス
my $temp_register_nc = << 'END';
	//-----------------------------------------------------------------------
	TP_ADD_LOG_REG_MEMBER(global.###FUNC_NAME###);
	###VAR_NAME### = Create_tjse_NativeClassClass_###FUNC_NAME###();
	val = tTJSVariant(###VAR_NAME###);
	###VAR_NAME###->Release();
	global->PropSet(TJS_MEMBERENSURE ###CALL_FLAG###, TJS_W("###FUNC_NAME###"), NULL, &val, global);
	//-----------------------------------------------------------------------
END



# クラス定義取得
my $temp_register_ico = << 'END';
	// ###ICO_CLASS### オブジェクト（クラス定義）を取得
	global->PropGet(TJS_MEMBERMUSTEXIST, TJS_W("###ICO_CLASS###"), NULL, &ico_result, global);
	iTJSDispatch2 * classdef_###ICO_CLASS### = ico_result.AsObjectNoAddRef();
END

my $temp_release = << 'END';
		TP_ADD_LOG_UNREG_MEMBER(global.###FUNC_NAME###);
		global->DeleteMember(0, TJS_W("###FUNC_NAME###"), NULL, global);
END

my $temp_release_wc = << 'END';
		if (classdef_###ICO_CLASS###) {
			TP_ADD_LOG_UNREG_MEMBER(global.###ICO_CLASS###.###FUNC_NAME###);
			classdef_###ICO_CLASS###->DeleteMember(0, TJS_W("###FUNC_NAME###"), NULL, classdef_###ICO_CLASS###);
		}
END

my @Data = ();


while (<STDIN>) {
	chomp;
	push @Data, $_;
}

foreach (@Data) {
	$_  =~ s/^\s+//;
	next if (!$_ || $_ =~ /^#/);
	last if ($_ =~ /^::/);
	my $f = $_;
	my $cppclass = '';
	my $cf = '';

	if ($f) {
		my $cppvar = "";
		my ($proprw, $propval);
		my $isprop = 0;

		# スタティック
		if ($f =~ s/\@static//) {
			$cf = '| TJS_STATICMEMBER';
		}

		# プロパティ
		if ($f =~ s/:([rw]+):(.*)//) {
			($proprw, $propval) = ($1, $2);
			$isprop = 1;
		}

		my $t;
		my $ico = '';
		my $classfile = '';

		# property
		if ($isprop) {
			$REGISTER_PROP .= qq|\tTP_AddConstRegisterExpression(propRegExpression, ttstr("${f}"), ${propval});\n|;
			$RELEASE_PROP  .= qq|\tTP_AddConstDeleteExpression(propDelExpression, ttstr("${f}"));\n|;

		# class def.
		} elsif ($f =~ s/\@class//) {
			if ($f =~ s/:(.+)//) {
				$classfile = $1;
			}
			$t = $temp_register_nc;

			$cppclass = getCppClassName($f);
			$cppvar = "v${cppclass}Instance";

			$t =~ s/###ICO_CLASS###/$ico/gs;
			$t =~ s/###FUNC_NAME###/$f/gs;
			$t =~ s/###VAR_NAME###/$cppvar/gs;
			$t =~ s/###CALL_FLAG###/$cf/gs;

			# クラスの時はiTJSDispatch2でいい
			$VARS  .= "iTJSDispatch2 * $cppvar;\n";
			$FUNCS .= "iTJSDispatch2 * Create_tjse_NativeClassClass_${f}();\n";	# クラスクリエイター

			# クラス登録
			push @RegisterList, $t; # 登録はリストに

			if ($classfile) {
				my $cdef = genClassdef($classfile, $f);
				push @RegisterList, $cdef;
			}

		# function
		} else {
			{
				# コンテキスト指定？
				my ($a, $b) = split(/\./, $f, 2);
				if ($b) {
					($ico, $f) = ($a, $b);
				}
			}

			$cppclass = getCppClassName(($ico ? "$ico." : '') . $f);
			$cppvar = "v${cppclass}Instance";

			if ($ico && !exists($ICO_CLASS{$ico})) {
				$t = $temp_register_ico;
				$t =~ s/###ICO_CLASS###/$ico/gs;
				push @RegisterList, $t; # 登録はリストに
				$RELEASE .= $t; # 解放は直に

				$ICO_CLASS{$ico} = 1;
			}

			$VARS .= "$cppclass * $cppvar;\n";

			$t = $ico ? $temp_register_wc : $temp_register;
			$t =~ s/###ICO_CLASS###/$ico/gs;
			$t =~ s/###CLASS_NAME###/$cppclass/gs;
			$t =~ s/###FUNC_NAME###/$f/gs;
			$t =~ s/###VAR_NAME###/$cppvar/gs;
			$t =~ s/###CALL_FLAG###/$cf/gs;
			push @RegisterList, $t; # 登録はリストに

		}
		$t = $ico ? $temp_release_wc : $temp_release;
		$t =~ s/###ICO_CLASS###/$ico/gs;
		$t =~ s/###FUNC_NAME###/$f/gs;
		push @ReleaseList, $t; # 解放もリストに

	}
}

$REGISTER .= join('', @RegisterList);
$RELEASE  .= join('', reverse @ReleaseList); # 解放は逆順



my $temp_base = << "END";
//---------------------------------------------------------------------------
// Generated by $0
//---------------------------------------------------------------------------
#include "${HNAME}"

// PP-macro TP_PLUGIN_NAME

static bool TPVerboseMode = false;

#ifndef TP_PLUGIN_NAME
#error マクロ TP_PLUGIN_NAME が設定されていない。コンパイラオプションにて -DTP_PLUGIN_NAME=\\"util_hoge\\" のように設定すべし。
#endif

#define TP_ADD_LOG_REG_MEMBER(_member)   TPAddLog_RegMember(ttstr(#_member));
#define TP_ADD_LOG_UNREG_MEMBER(_member) TPAddLog_UnregMember(ttstr(#_member));
#define TP_ADD_LOG_LINK_EVENT(_func, _ev)   TPAddLog_LinkEvent(ttstr(#_func), ttstr(#_ev));

END

# use holder ?
if ($USE_DH) {
	$temp_base .= << 'END';
#include "TableHolder.h"
#include <lmcons.h>

//---------------------------------------------------------------------------

CTableHolder * TableHolder;

END
} else {
	$temp_base .= << 'END';
#include <lmcons.h>

//---------------------------------------------------------------------------

END
}

# リンク・アンリンクイベント
foreach my $func qw(V2Link V2Unlink) {
	foreach my $ev qw(Enter Exit) {
		$temp_base .= << "END";

// default impl.
static void TPOn${func}${ev}_Default(void)
{
	return;
}

// set default
static void (*TPOn${func}${ev})(void) = TPOn${func}${ev}_Default;

// override
void TPSet${func}${ev}Callback(void (*callback)(void))
{
	TPOn${func}${ev} = callback;
}
END
	}
}

$temp_base .= << "END";

static void TPAddLog_RegMember(const ttstr & member)
{
	if (TPVerboseMode) {
		ttstr msg = ttstr(TP_PLUGIN_NAME) + TJS_W(": Add member ") + member + TJS_W(" ...");
		TVPAddLog(msg);
	}
}

static void TPAddLog_UnregMember(const ttstr & member)
{
	if (TPVerboseMode) {
		ttstr msg = ttstr(TP_PLUGIN_NAME) + TJS_W(": Delete member ") + member + TJS_W(" ...");
		TVPAddLog(msg);
	}
}

static void TPAddLog_LinkEvent(const ttstr & func, const ttstr & ev)
{
	if (TPVerboseMode) {
		ttstr msg = ttstr(TP_PLUGIN_NAME) + TJS_W(": on ") + func + TJS_W(" ") + ev + TJS_W(" ...");
		TVPAddLog(msg);
	}
}

// Object creators
${FUNCS}

// TJS2 objects
${VARS}

//---------------------------------------------------------------------------
#pragma argsused
int WINAPI DllEntryPoint(HINSTANCE hinst, unsigned long reason, void* lpReserved)
{
	TPSetLinkEventCallback();

	return 1;
}
//---------------------------------------------------------------------------
static tjs_int GlobalRefCountAtInit = 0;
extern "C" HRESULT _stdcall _export V2Link(iTVPFunctionExporter *exporter)
{
	TVPInitImportStub(exporter);

	tTJSVariant val;
	if(TVPGetCommandLine(TJS_W("-${PluginOption1}"), &val))
	{
		TPVerboseMode = (val == ttstr("yes"));
	}

	tTJSVariant ico_result;
	iTJSDispatch2 * global = TVPGetScriptDispatch();

	TP_ADD_LOG_LINK_EVENT(V2Link, Enter);
	TPOnV2LinkEnter();

END

# use holder ?
if ($USE_DH) {
	$temp_base .= << "END";
	//-----------------------------------------------------------------------
	// データホルダー
	TableHolder = new CTableHolder();
	TableHolder->createTable(DEFAULT_ANGLE_ACC, DEFAULT_SIN_ACC);
	//-----------------------------------------------------------------------
END
}

$temp_base .= << "END";
${REGISTER}

	ttstr propRegExpression;
${REGISTER_PROP}
	TVPExecuteExpression(propRegExpression);

	global->Release();
	val.Clear();
	GlobalRefCountAtInit = TVPPluginGlobalRefCount;

	TP_ADD_LOG_LINK_EVENT(V2Link, Exit);
	TPOnV2LinkExit();

	return S_OK;
}
//---------------------------------------------------------------------------
extern "C" HRESULT _stdcall _export V2Unlink()
{
	if(TVPPluginGlobalRefCount > GlobalRefCountAtInit) return E_FAIL;

	TP_ADD_LOG_LINK_EVENT(V2Unlink, Enter);
	TPOnV2UnlinkEnter();

	iTJSDispatch2 * global = TVPGetScriptDispatch();
	tTJSVariant ico_result;

	if (global) {
	ttstr propDelExpression;
${RELEASE_PROP}
	TVPExecuteExpression(propDelExpression);

${RELEASE}

END

# use holder ?
if ($USE_DH) {
	$temp_base .= << "END";
		// データホルダー
		delete TableHolder;
END
}


$temp_base .= << "END";
	}

	TP_ADD_LOG_LINK_EVENT(V2Unlink, Exit);
	TPOnV2UnlinkExit();

	if (global) global->Release();
	TVPUninitImportStub();

	return S_OK;
}
//---------------------------------------------------------------------------
static tjs_char option_message[] = 
L"プラグインモード:${HNAME_BASE} Verbose Mode;${HNAME_BASE}.dllのモードを設定します。|"
L"${PluginOption1}|select,*no;通常モードです。,yes;Link/Unlink時にログを出力します。\\n";
//---------------------------------------------------------------------------
extern "C" const wchar_t* _stdcall _export GetOptionDesc()
{
	return option_message;
}
//---------------------------------------------------------------------------

#undef TP_ADD_LOG_REG_MEMBER
#undef TP_ADD_LOG_UNREG_MEMBER

END

print $temp_base;



sub genClassdef {
	my ($file, $name) = @_;
	my $tjs = '';
	my $ud = 0;

	open (FT, $file) or die ("$file: $!");
	binmode(FT, $ENCODE_NAME);
	while (<FT>) {
		chomp;
		my $x = $_;
		$tjs .= qq|\t\tTJS_W("$x\\n")\n|;
	}
	close FT;
	my $rv = << "END";
	//-----------------------------------------------------------------------
	// Class $name extention
	TVPExecuteExpression(
$tjs		, NULL);
	//-----------------------------------------------------------------------
END

	return $rv;
}


