
# Original code: written by W.Dee
# Modified by mey

my $ProcessRoot = '..';
my $RevisionHeader = 'common/TPRevision.h';
my $RevisionMake   = 'revision.mak';
my $HeaderID = '_TVP_PLUGIN_REVISION_H_';

undef $/;


# usage: $0  #.#.#.#level  "TVP version"
my ($TPVersion, $TVPVersion) = ($ARGV[0] || 'unknown', $ARGV[1] || 'not specified');


$newest_rev = 0;

sub check_revision
{
	local($fn, $content, $entry);
	$fn = $_[0];
	open FH, "$fn";
	$content = <FH>;
	while($content =~ /\<entry.*?revision=\"(\d+)\".*?\/\>/gs)
	{
		;#"
		$newest_rev = $1 if($newest_rev < $1);
	}
}


sub process_dir
{
	local($dir, @dirlist, $file, $w, $h, $fn, $newfn, $content, $ext, $md5);
	$dir = $_[0];


	if(opendir(DH, "$dir"))
	{
		@dirlist = readdir DH;
		closedir(DH);
		foreach $file (@dirlist)
		{
			if($file eq '.svn')
			{
				;# a svn meta-information directory
				$fn = "$dir/$file/entries";
				&check_revision($fn);
			}
			elsif($file !~ /^\./)
			{
				$fn = "$dir/$file";
				if(!&process_dir($fn))
				{
					;# a file
				}
			}
		}
		return 1;
	}
	return 0;
}

process_dir($ProcessRoot);

if ($newest_rev) {
	$TPVersion =~ s/^(\d+\.\d+\.\d+)\.\d+(.*)/$1.$newest_rev$2/;
} else {
	$TPVersion =~ /^\d+\.\d+\.\d+\.(\d+)/;
	$newest_rev = $1 || 0;
}

open FH, "$ProcessRoot/.svn/entries";

$revision = "<unknown svn revision>";

$lines = <FH>;

if($lines =~ /\<entry(.*?)name=\"\"(.*?)\/\>/s)
{
	;# "
	$lines = "$1\t$2";

	$u = $1 if($lines =~ /url=\"(.*?)\"/s);
	;# "

	if($u ne "")
	{
		$revision = "SVN rev=$newest_rev url=$u";
	}
}


open FH, ">$RevisionHeader";

print FH << "EOF";
/* this file is automatically updated by $0. */
#ifndef $HeaderID
#define $HeaderID

#define TVP_PLUGIN_VERSION_NUMBER "$TPVersion"
#define TVP_PLUGIN_TVP_VERSION_NUMBER "$TVPVersion"
#define TVP_PLUGIN_SVN_REVISION "$revision"
#define TVP_PLUGIN_SVN_REVISION_NUMBER "$newest_rev"

#endif  $HeaderID
EOF
close(FH);

open FH, ">$RevisionMake";

print FH << "EOF";
# this file is automatically updated by $0.
TP_VERSION_BUILD   = $newest_rev
EOF
close(FH);



print << "END";
TVP Gottani Plugin ver. $TPVersion for TVP ver. $TVPVersion
$revision
END

